# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Uses the Azure Python SDK to collect and return details related to Azure."""
from typing import Dict, List

import pandas as pd
import requests
from azure.common.exceptions import CloudError

from .azure_data import AzureData
from ..common.exceptions import MsticpyAzureConfigError
from ..common.wsconfig import WorkspaceConfig

_PATH_MAPPING = {
    "ops_path": "/providers/Microsoft.SecurityInsights/operations",
    "alert_rules": "/providers/Microsoft.SecurityInsights/alertRules",
    "ss_path": "/savedSearches",
    "bookmarks": "/providers/Microsoft.SecurityInsights/bookmarks",
}


class AzureSentinel(AzureData):
    """Class for reutrning key Azure Sentinel elements."""

    def __init__(self, connect: bool = False):
        """Initialize connector for Azure APIs."""
        super().__init__()
        self.config = None

    def connect(
        self,
        auth_methods: List = None,
        silent: bool = False,
    ):
        """Authenticate with the SDK & API."""
        super().connect(auth_methods=auth_methods, silent=silent)
        self.token = _get_token(self.credentials)
        self.res_group_url = None
        self.prov_path = None

    def get_sentinel_workspaces(self, sub_id: str = None) -> Dict:
        """
        Return a list of Azure Sentinel workspaces in a Subscription.

        Parameters
        ----------
        sub_id : str
            The subscription ID to get a list of workspaces from.
            If not provided it will attempt to get sub_id from config files.

        Returns
        -------
        Dict
            A dictionary of workspace names and ids

        """
        # If a subscription ID isn't provided try and get one from config files.
        if not sub_id:
            config = self._check_config(["subscription_id"])
            sub_id = config["subscription_id"]

        print("Finding Azure Sentinel Workspaces...")
        res = self.get_resources(sub_id=sub_id)  # type: ignore
        # handle no results
        if isinstance(res, pd.DataFrame) and not res.empty:
            sentinel = res[
                (res["resource_type"] == "Microsoft.OperationsManagement/solutions")
                & (res["name"].str.startswith("SecurityInsights"))
            ]
            workspaces = []
            for wrkspace in sentinel["resource_id"]:
                res_details = self.get_resource_details(
                    sub_id=sub_id, resource_id=wrkspace  # type: ignore
                )
                workspaces.append(res_details["properties"]["workspaceResourceId"])

            workspaces_dict = {}
            for wrkspace in workspaces:
                name = wrkspace.split("/")[-1]
                workspaces_dict.update({name: wrkspace})

            return workspaces_dict

        print(f"No Azure Sentinel workspaces in {sub_id}")
        return {}

    def get_hunting_queries(
        self,
        res_id: str = None,
        sub_id: str = None,
        res_grp: str = None,
        ws_name: str = None,
    ) -> pd.DataFrame:
        """
        Return all hunting queries in an Azure Sentinel workspace.

        Parameters
        ----------
        res_id : str, optional
            Resource ID of the workspace, if not provided details from config file will be used.
        sub_id : str, optional
            Sub ID of the workspace, to be used if not providing Resource ID.
        res_grp : str, optional
            Resource Group name of the workspace, to be used if not providing Resource ID.
        ws_name : str, optional
            Workspace name of the workspace, to be used if not providing Resource ID.

        Returns
        -------
        pd.DataFrame
            A table of the hunting queries.

        """
        # If res_id isn't provided try and get them from config
        if not res_id:
            if not sub_id or not res_grp or not ws_name:
                config = self._check_config(
                    ["subscription_id", "resource_group", "workspace_name"]
                )
                sub_id = config["subscription_id"]
                res_grp = config["resource_group"]
                ws_name = config["workspace_name"]
            res_id = f"/subscriptions/{sub_id}/resourcegroups/{res_grp}"
            res_id = (
                res_id + "/providers/Microsoft.OperationalInsights/workspaces/{ws_name}"
            )

        url = _build_paths(res_id)
        saved_searches_url = url + _PATH_MAPPING["ss_path"]
        params = {"api-version": "2017-04-26-preview"}

        response = requests.get(
            saved_searches_url, headers=_get_api_headers(self.token), params=params
        )
        if response.status_code == 200:
            queries_df = _azs_api_result_to_df(response)
        else:
            raise CloudError("Could not get alert rules.")

        return queries_df[queries_df["Category"] == "Hunting Queries"]

    def get_alert_rules(
        self,
        res_id: str = None,
        sub_id: str = None,
        res_grp: str = None,
        ws_name: str = None,
    ) -> pd.DataFrame:
        """
        Return all Azure Sentinel alert rules for a workspace.

        Parameters
        ----------
        res_id : str, optional
            Resource ID of the workspace, if not provided details from config file will be used.
        sub_id : str, optional
            Sub ID of the workspace, to be used if not providing Resource ID.
        res_grp : str, optional
            Resource Group name of the workspace, to be used if not providing Resource ID.
        ws_name : str, optional
            Workspace name of the workspace, to be used if not providing Resource ID.

        Returns
        -------
        pd.DataFrame
            A table of the workspace's alert rules.

        """
        if not res_id:
            if not sub_id or not res_grp or not ws_name:
                config = self._check_config(
                    ["subscription_id", "resource_group", "workspace_name"]
                )
                sub_id = config["subscription_id"]
                res_grp = config["resource_group"]
                ws_name = config["workspace_name"]
            res_id = f"/subscriptions/{sub_id}/resourcegroups/{res_grp}"
            res_id = (
                res_id + "/providers/Microsoft.OperationalInsights/workspaces/{ws_name}"
            )

        url = _build_paths(res_id)
        alert_rules_url = url + _PATH_MAPPING["alert_rules"]
        params = {"api-version": "2020-01-01"}

        response = requests.get(
            alert_rules_url, headers=_get_api_headers(self.token), params=params
        )
        if response.status_code == 200:
            alerts_df = _azs_api_result_to_df(response)
        else:
            raise CloudError("Could not get alert rules.")

        return alerts_df

    def get_bookmarks(
        self,
        res_id: str = None,
        sub_id: str = None,
        res_grp: str = None,
        ws_name: str = None,
    ) -> pd.DataFrame:
        """
        Return a list of Bookmarks from a Sentinel workspace.

        Parameters
        ----------
        res_id : str, optional
            Resource ID of the workspace, if not provided details from config file will be used.
        sub_id : str, optional
            Sub ID of the workspace, to be used if not providing Resource ID.
        res_grp : str, optional
            Resource Group name of the workspace, to be used if not providing Resource ID.
        ws_name : str, optional
            Workspace name of the workspace, to be used if not providing Resource ID.

        Returns
        -------
        pd.DataFrame
            A set of bookmarks.

        Raises
        ------
        CloudError
            If bookmark collection fails.

        """
        if not res_id:
            if not sub_id or not res_grp or not ws_name:
                config = self._check_config(
                    ["subscription_id", "resource_group", "workspace_name"]
                )
                sub_id = config["subscription_id"]
                res_grp = config["resource_group"]
                ws_name = config["workspace_name"]
            res_id = f"/subscriptions/{sub_id}/resourcegroups/{res_grp}"
            res_id = (
                res_id + "/providers/Microsoft.OperationalInsights/workspaces/{ws_name}"
            )

        url = _build_paths(res_id)
        bookmarks_url = url + _PATH_MAPPING["bookmarks"]
        params = {"api-version": "2020-01-01"}

        response = requests.get(
            bookmarks_url, headers=_get_api_headers(self.token), params=params
        )
        if response.status_code == 200:
            bookmarks_df = _azs_api_result_to_df(response)
        else:
            raise CloudError("Could not get bookmarks.")

        return bookmarks_df

    def _check_config(self, items: List) -> Dict:
        """
        Get parameters from default config files.

        Parameters
        ----------
        items : List
            The items to get from the config.

        Returns
        -------
        Dict
            The config items.

        """
        config_items = {}
        if not self.config:
            self.config = WorkspaceConfig()  # type: ignore
        for item in items:
            if item in self.config:  # type: ignore
                config_items.update({item: self.config[item]})  # type: ignore
            else:
                raise MsticpyAzureConfigError(f"No {item} avaliable in config.")

        return config_items


def _build_paths(resid) -> str:
    """Build a API URL from an Azure resource ID."""
    res_info = {
        "subscription_id": resid.split("/")[2],
        "resource_group": resid.split("/")[4],
        "workspace_name": resid.split("/")[-1],
    }

    url_part1 = (
        f"https://management.azure.com/subscriptions/{res_info['subscription_id']}"
    )
    url_part2 = f"/resourceGroups/{res_info['resource_group']}"
    url_part3 = f"/providers/Microsoft.OperationalInsights/workspaces/{res_info['workspace_name']}"

    return url_part1 + url_part2 + url_part3


def _get_token(credential) -> str:
    """Extract token from a azure.identity object."""
    token = credential.modern.get_token("https://management.azure.com/.default")
    return token.token


def _get_api_headers(token):
    """Return authorization header with current token."""
    return {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }


def _azs_api_result_to_df(response) -> pd.DataFrame:
    """Convert API reponse to a Pandas dataframe."""
    j_resp = response.json()
    if response.status_code != 200 or not j_resp or "value" not in j_resp:
        raise ValueError("No valid JSON result in response")
    queries_raw_df = pd.DataFrame(j_resp["value"])
    query_props_df = pd.json_normalize(queries_raw_df["properties"])
    return pd.concat([queries_raw_df, query_props_df], axis=1).drop(
        columns="properties"
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data provider loader."""
import warnings
from datetime import datetime
from functools import partial
from itertools import tee
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union

import pandas as pd
from tqdm.auto import tqdm

from .._version import VERSION
from ..common import pkg_config as config
from ..common.utility import export, valid_pyname
from .browsers.query_browser import browse_queries
from .drivers import (
    DriverBase,
    KqlDriver,
    LocalDataDriver,
    MDATPDriver,
    MordorDriver,
    SecurityGraphDriver,
    SplunkDriver,
)
from .param_extractor import extract_query_params
from .query_container import QueryContainer
from .query_defns import DataEnvironment
from .query_source import QuerySource
from .query_store import QueryStore

__version__ = VERSION
__author__ = "Ian Hellen"

_PROVIDER_DIR = "providers"

_ENVIRONMENT_DRIVERS = {
    DataEnvironment.LogAnalytics: KqlDriver,
    DataEnvironment.AzureSecurityCenter: KqlDriver,
    DataEnvironment.SecurityGraph: SecurityGraphDriver,
    DataEnvironment.MDATP: MDATPDriver,
    DataEnvironment.LocalData: LocalDataDriver,
    DataEnvironment.Splunk: SplunkDriver,
    DataEnvironment.Mordor: MordorDriver,
}


@export
class QueryProvider:
    """
    Container for query store and query execution provider.

    Instances of this class hold the query set and execution
    methods for a specific data environment.

    """

    def __init__(  # noqa: MC0001
        self,
        data_environment: Union[str, DataEnvironment],
        driver: DriverBase = None,
        query_paths: List[str] = None,
        **kwargs,
    ):
        """
        Query provider interface to queries.

        Parameters
        ----------
        data_environment : Union[str, DataEnvironment]
            Name or Enum of environment for the QueryProvider
        driver : DriverBase, optional
            Override the builtin driver (query execution class)
            and use your own driver (must inherit from
            `DriverBase`)
        query_paths : List[str]
            Additional paths to look for query definitions.
        kwargs :
            Other arguments are passed to the data provider driver.

        See Also
        --------
        DataProviderBase : base class for data query providers.

        """
        if isinstance(data_environment, str):
            data_env = DataEnvironment.parse(data_environment)
            if data_env != DataEnvironment.Unknown:
                data_environment = data_env
            else:
                raise TypeError(f"Unknown data environment {data_environment}")

        self._environment = data_environment.name

        if driver is None:
            driver_class = _ENVIRONMENT_DRIVERS[data_environment]
            if issubclass(driver_class, DriverBase):
                driver = driver_class(**kwargs)  # type: ignore
            else:
                raise LookupError(
                    "Could not find suitable data provider for", f" {self._environment}"
                )

        self._query_provider = driver
        self.all_queries = QueryContainer()

        # Add any query files
        data_env_queries: Dict[str, QueryStore] = {}
        if driver.use_query_paths:
            data_env_queries.update(
                self._read_queries_from_paths(query_paths=query_paths)
            )
        self._query_store = data_env_queries.get(
            self._environment, QueryStore(self._environment)
        )
        self._add_query_functions()

    def __getattr__(self, name):
        """Return the value of the named property 'name'."""
        if "." in name:
            parent_name, child_name = name.split(".", maxsplit=1)
            parent = getattr(self, parent_name, None)
            if parent:
                return getattr(parent, child_name)
        raise AttributeError(f"{name} is not a valid attribute.")

    def connect(self, connection_str: str = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : str
            Connection string for the data source

        """
        self._query_provider.connect(connection_str=connection_str, **kwargs)

        # If the driver has any attributes to expose via the provider
        # add those here.
        for attr_name, attr in self._query_provider.public_attribs.items():
            setattr(self, attr_name, attr)

        # Add any built-in or dynamically retrieved queries from driver
        if self._query_provider.has_driver_queries:
            driver_queries = self._query_provider.driver_queries
            self._add_driver_queries(queries=driver_queries)

    @property
    def connected(self) -> bool:
        """
        Return True if the provider is connected.

        Returns
        -------
        bool
            True if the provider is connected.

        """
        return self._query_provider.connected

    @property
    def connection_string(self) -> str:
        """
        Return provider connection string.

        Returns
        -------
        str
            Provider connection string.

        """
        return self._query_provider.current_connection

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return self._query_provider.schema

    @property
    def schema_tables(self) -> List[str]:
        """
        Return list of tables in the data schema of the connection.

        Returns
        -------
        List[str]
            Tables in the of current connection.

        """
        return list(self._query_provider.schema.keys())

    def import_query_file(self, query_file: str):
        """
        Import a yaml data source definition.

        Parameters
        ----------
        query_file : str
            Path to the file to import

        """
        self._query_store.import_file(query_file)
        self._add_query_functions()

    @classmethod
    def list_data_environments(cls) -> List[str]:
        """
        Return list of current data environments.

        Returns
        -------
        List[str]
            List of current data environments

        """
        return [env for env in DataEnvironment.__members__ if env != "Unknown"]

    def list_queries(self) -> List[str]:
        """
        Return list of family.query in the store.

        Returns
        -------
        Iterable[str]
            List of queries

        """
        return list(self._query_store.query_names)

    def query_help(self, query_name):
        """Print help for query."""
        self._query_store[query_name].help()

    def get_query(self, query_name) -> str:
        """Return the raw query text."""
        return self._query_store[query_name].query

    def exec_query(self, query: str, **kwargs) -> Union[pd.DataFrame, Any]:
        """
        Execute simple query string.

        Parameters
        ----------
        query : str
            [description]

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results - a DataFrame if successful
            or a KqlResult if unsuccessful.

        """
        query_options = kwargs.pop("query_options", {}) or kwargs
        return self._query_provider.query(query, **query_options)

    def browse_queries(self, **kwargs):
        """
        Return QueryProvider query browser.

        Parameters
        ----------
        query_provider : QueryProvider
            Initialized query provider.

        Other Parameters
        ----------------
        kwargs :
            passed to SelectItem constructor.

        Returns
        -------
        SelectItem
            SelectItem browser for TI Data.

        """
        return browse_queries(self, **kwargs)

    def _execute_query(self, *args, **kwargs) -> Union[pd.DataFrame, Any]:
        if not self._query_provider.loaded:
            raise ValueError("Provider is not loaded.")
        if not self._query_provider.connected:
            raise ValueError(
                "No connection to a data source.",
                "Please call connect(connection_str) and retry.",
            )
        query_name = kwargs.pop("query_name")
        family = kwargs.pop("query_path")

        query_source = self._query_store.get_query(
            query_path=family, query_name=query_name
        )
        if "help" in args or "?" in args:
            query_source.help()
            return None

        params, missing = extract_query_params(query_source, *args, **kwargs)
        if missing:
            query_source.help()
            raise ValueError(f"No values found for these parameters: {missing}")

        split_by = kwargs.pop("split_query_by", None)
        if split_by:
            split_result = self._exec_split_query(
                split_by=split_by,
                query_source=query_source,
                query_params=params,
                args=args,
                **kwargs,
            )
            if split_result is not None:
                return split_result
            # if split queries could not be created, fall back to default
        query_str = query_source.create_query(
            formatters=self._query_provider.formatters, **params
        )
        if "print" in args or "query" in args:
            return query_str

        # Handle any query options passed
        query_options = self._get_query_options(params, kwargs)
        return self._query_provider.query(query_str, query_source, **query_options)

    @staticmethod
    def _get_query_options(
        params: Dict[str, Any], kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Return any kwargs not already in params."""
        query_options = kwargs.pop("query_options", {})
        if not query_options:
            # Any kwargs left over we send to the query provider driver
            query_options = {
                key: val for key, val in kwargs.items() if key not in params
            }
        return query_options

    def _read_queries_from_paths(self, query_paths) -> Dict[str, QueryStore]:
        """Fetch queries from YAML files in specified paths."""
        settings: Dict[str, Any] = config.settings.get(  # type: ignore
            "QueryDefinitions"
        )  # type: ignore
        all_query_paths = []
        for default_path in settings.get("Default"):  # type: ignore
            qry_path = self._resolve_package_path(default_path)
            if qry_path:
                all_query_paths.append(qry_path)

        if settings.get("Custom") is not None:
            for custom_path in settings.get("Custom"):  # type: ignore
                qry_path = self._resolve_path(custom_path)
                if qry_path:
                    all_query_paths.append(qry_path)
        if query_paths:
            for custom_path in query_paths:
                qry_path = self._resolve_path(custom_path)
                if qry_path:
                    all_query_paths.append(qry_path)

        if not all_query_paths:
            raise RuntimeError(
                "No valid query definition files found. ",
                "Please check your msticpyconfig.yaml settings.",
            )
        return QueryStore.import_files(source_path=all_query_paths, recursive=True)

    def _add_query_functions(self):
        """Add queries to the module as callable methods."""
        for qual_query_name in self.list_queries():
            query_path = qual_query_name.split(".")
            query_name = query_path[-1]
            current_node = self
            for container_name in query_path[:-1]:
                container_name = valid_pyname(container_name)
                if hasattr(current_node, container_name):
                    current_node = getattr(current_node, container_name)
                else:
                    new_node = QueryContainer()
                    setattr(current_node, container_name, new_node)
                    current_node = new_node

            query_cont_name = ".".join(query_path[:-1])

            # Create the partial function
            query_func = partial(
                self._execute_query, query_path=query_cont_name, query_name=query_name
            )
            query_func.__doc__ = self._query_store.get_query(
                query_path=query_cont_name, query_name=query_name
            ).create_doc_string()

            query_name = valid_pyname(query_name)
            setattr(current_node, query_name, query_func)
            setattr(self.all_queries, query_name, query_func)

    def _add_driver_queries(self, queries: Iterable[Dict[str, str]]):
        """Add driver queries to the query store."""
        for query in queries:
            self._query_store.add_query(
                name=query["name"],
                query=query["query"],
                query_paths=query["query_container"],
                description=query["description"],
            )
        # For now, just add all of the functions again (with any connect-time acquired
        # queries) - we could be more efficient than this but unless there are 1000s of
        # queries it should not be noticeable.
        self._add_query_functions()

    def _exec_split_query(
        self,
        split_by: str,
        query_source: QuerySource,
        query_params: Dict[str, Any],
        args,
        **kwargs,
    ) -> Union[pd.DataFrame, str, None]:
        start = query_params.pop("start", None)
        end = query_params.pop("end", None)
        if not (start or end):
            print(
                "Cannot split a query that does not have 'start' and 'end' parameters"
            )
            return None
        try:
            split_delta = pd.Timedelta(split_by)
        except ValueError:
            split_delta = pd.Timedelta("1D")

        ranges = self._calc_split_ranges(start, end, split_delta)

        split_queries = [
            query_source.create_query(
                formatters=self._query_provider.formatters,
                start=q_start,
                end=q_end,
                **query_params,
            )
            for q_start, q_end in ranges
        ]
        if "print" in args or "query" in args:
            return "\n\n".join(split_queries)

        # Retrive any query options passed (other than query params)
        # and send to query function.
        query_options = self._get_query_options(query_params, kwargs)
        query_dfs = [
            self._query_provider.query(query_str, query_source, **query_options)
            for query_str in tqdm(split_queries, unit="sub-queries", desc="Running")
        ]

        return pd.concat(query_dfs)

    @staticmethod
    def _calc_split_ranges(start: datetime, end: datetime, split_delta: pd.Timedelta):
        """Return a list of time ranges split by `split_delta`."""
        # Use pandas date_range and split the result into 2 iterables
        s_ranges, e_ranges = tee(pd.date_range(start, end, freq=split_delta))
        next(e_ranges, None)  # skip to the next item in the 2nd iterable
        # Zip them together to get a list of (start, end) tuples of ranges
        # Note: we subtract 1 nanosecond from the 'end' value of each range so
        # to avoid getting duplicated records at the boundaries of the ranges.
        # Some providers don't have nanosecond granularity so we might
        # get duplicates in these cases
        ranges = [
            (s_time, e_time - pd.Timedelta("1ns"))
            for s_time, e_time in zip(s_ranges, e_ranges)
        ]

        # Since the generated time ranges are based on deltas from 'start'
        # we need to adjust the end time on the final range.
        # If the difference between the calculated last range end and
        # the query 'end' that the user requested is small (< 10% of a delta),
        # we just replace the last "end" time with our query end time.
        if (ranges[-1][1] - end) < (split_delta / 10):
            ranges[-1] = ranges[-1][0], end
        else:
            # otherwise append a new range starting after the last range
            # in ranges and ending in 'end"
            # note - we need to add back our subtracted 1 nanosecond
            ranges.append((ranges[-1][0] + pd.Timedelta("1ns"), end))
        return ranges

    @classmethod
    def _resolve_package_path(cls, config_path: str) -> Optional[str]:
        """Resolve path relative to current package."""
        if not Path(config_path).is_absolute():
            config_path = str(Path(__file__).resolve().parent.joinpath(config_path))
        if not Path(config_path).is_dir():
            warnings.warn(f"Custom query definitions path {config_path} not found")
            return None
        return config_path

    @classmethod
    def _resolve_path(cls, config_path: str) -> Optional[str]:
        """Resolve path."""
        if not Path(config_path).is_absolute():
            config_path = str(Path(config_path).resolve())
        if not Path(config_path).is_dir():
            warnings.warn(f"Custom query definitions path {config_path} not found")
            return None
        return config_path



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data provider sub-package."""
# flake8: noqa: F403
from .driver_base import DriverBase
from .kql_driver import KqlDriver
from .security_graph_driver import SecurityGraphDriver
from .mdatp_driver import MDATPDriver
from .local_data_driver import LocalDataDriver
from .splunk_driver import SplunkDriver
from .mordor_driver import MordorDriver

from ..._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""QueryStore class - holds a collection of QuerySources."""
from collections import defaultdict
from os import path
from typing import Any, Dict, Iterable, Set, Union, Optional, List

from .._version import VERSION
from .query_defns import DataEnvironment, DataFamily
from .data_query_reader import find_yaml_files, read_query_def_file
from .query_source import QuerySource

__version__ = VERSION
__author__ = "Ian Hellen"


def _get_dot_path(elem_path: str, data_map: dict) -> Any:
    """
    Return For dotted attribute, tries to search.

    Parameters
    ----------
    elem_path : str
        The attribute name or prefix.name
    data_map : dict
        The dictionary/map to search through.

    Returns
    -------
    Any
        The attribute value

    Raises
    ------
    KeyError
        If the key/subkey is not found

    """
    # if this is directly in the map return it
    if elem_path in data_map:
        return data_map[elem_path]
    # otherwise partition into prefix and name
    prefix, _, name = elem_path.rpartition(".")
    attrib = data_map.get(prefix)
    if isinstance(attrib, dict) and name in attrib:
        return attrib[name]
    raise KeyError(f"'{elem_path}' not found")


class QueryStore:
    """
    Repository for query definitions for a data environment.

    Attributes
    ----------
    environment: str
        The data environment for the queries.
    data_families: Dict[str, Dict[str, QuerySource]]
        The set of data families and associated queries
        for each.

    """

    def __init__(self, environment: str):
        """
        Intialize a QueryStore for a new environment.

        Parameters
        ----------
        environment : str
            The data environment

        """
        self.environment: str = environment
        self.data_families: Dict[str, Dict[str, QuerySource]] = defaultdict(dict)
        self.data_family_defaults: Dict[str, Dict[str, Any]] = defaultdict(dict)

    def __getattr__(self, name: str):
        """Return the item in dot-separated path `name`."""
        return _get_dot_path(elem_path=name, data_map=self.data_families)

    def __getitem__(self, key: str):
        """Allow query retrieval using dotted key path."""
        return _get_dot_path(elem_path=key, data_map=self.data_families)

    @property
    def query_names(self) -> Iterable[str]:
        """
        Return list of family.query in the store.

        Returns
        -------
        Iterable[str]
            List of queries

        """
        for family in sorted(self.data_families):

            yield from [
                f"{family}.{query}"
                for query in sorted(self.data_families[family].keys())
            ]

    def add_data_source(self, source: QuerySource):
        """
        Add a datasource/query to the store.

        Parameters
        ----------
        source : QuerySource
            The source to add. An existing item with
            the same name will be overwritten

        """
        source.query_store = self
        for family in source.data_families:
            self.data_families[family][source.name] = source
            # we want to update any new defaults for the data family
            self.data_family_defaults[family].update(source.defaults)

            # we also replace the defaults dict in each source with
            # a reference to the data family defaults so that as
            # new files are read in defaults are shared by all sources.
            source.defaults = self.data_family_defaults[family]

            valid, failures = source.validate()
            if not valid:
                raise ImportError(source.name, failures)

    def add_query(
        self,
        name: str,
        query: str,
        query_paths: Union[str, List[str]],
        description: str = None,
    ):
        """
        Add a query from name/query text.

        Parameters
        ----------
        name : str
            name of the query
        query : str
            The query string
        query_paths : Union[str, List[str]]
            The path/data_family to categorize.
            Multiple paths can be specified. If the path is dotted,
            this will cause the query to be displayed in the corresponding
            hierarchy.
        description : str, optional
            Query description

        """
        prefix = ""
        if "." in name:
            prefix, _, name = name.rpartition(".")

        if isinstance(query_paths, str):
            query_paths = [query_paths]
        if prefix:
            query_paths = [f"{q_path}.{prefix}" for q_path in query_paths]

        src_dict = {"args": {"query": query}, "description": description or name}
        md_dict = {"data_families": query_paths}

        query_source = QuerySource(
            name=name, source=src_dict, defaults={}, metadata=md_dict
        )
        self.add_data_source(query_source)

    def import_file(self, query_file: str):
        """
        Import a yaml data source definition.

        Parameters
        ----------
        query_file : str
            Path to the file to import

        Raises
        ------
        ImportError
            File read error or Syntax or semantic error found in
            the source file.

        """
        sources, defaults, metadata = read_query_def_file(query_file)

        for source_name, source in sources.items():
            new_source = QuerySource(source_name, source, defaults, metadata)
            self.add_data_source(new_source)

    @classmethod  # noqa: MC0001
    def import_files(
        cls, source_path: list, recursive: bool = False
    ) -> Dict[str, "QueryStore"]:
        """
        Import multiple query definition files from directory path.

        Parameters
        ----------
        source_path : str
            The folder containing the yaml definition files.
        recursive : bool, optional
            True to recurse sub-directories
            (the default is False, which only reads from the top level)

        Returns
        -------
        Dict[str, 'QueryStore']
            Dictionary of one or more environments and the
            QueryStore containing the queries for each environment.

        Raises
        ------
        FileNotFoundError
            File read error or Syntax or semantic error found in
            a source file.

        """
        env_stores: Dict[str, QueryStore] = dict()
        for query_dir in source_path:
            if not path.isdir(query_dir):
                raise FileNotFoundError(f"{query_dir} is not a directory")
            for file_path in find_yaml_files(query_dir, recursive):
                try:
                    sources, defaults, metadata = read_query_def_file(str(file_path))
                except ValueError:
                    print(
                        f"{file_path} is not a valid query definition file - skipping."
                    )
                    continue

                for env_value in metadata["data_environments"]:
                    if "." in env_value:
                        env_value = env_value.split(".")[1]
                    environment = DataEnvironment.parse(env_value)
                    if environment == DataEnvironment.Unknown:
                        raise ValueError(f"Unknown environment {env_value}")

                    if environment.name not in env_stores:
                        env_stores[environment.name] = cls(environment=environment.name)
                    for source_name, source in sources.items():
                        new_source = QuerySource(
                            source_name, source, defaults, metadata
                        )
                        env_stores[environment.name].add_data_source(new_source)

        return env_stores

    def get_query(
        self, query_name: str, query_path: Union[str, DataFamily] = None
    ) -> "QuerySource":
        """
        Return query with name `data_family` and `query_name`.

        Parameters
        ----------
        query_name: str
            Name of the query
        query_path: Union[str, DataFamily]
            The data family for the query

        Returns
        -------
        QuerySource
            Query matching name and family.

        """
        if query_path and isinstance(query_path, DataFamily):
            query_path = query_path.name
        if "." in query_name:
            query_parts = query_name.split(".")
            query_container = ".".join(query_parts[:-1])
            query_name = query_parts[-1]
            if query_container in self.data_families:
                query_path = query_container
            elif query_path:
                query_container = ".".join(
                    [query_path, query_container]  # type: ignore
                )
                if query_container in self.data_families:
                    query_path = query_container
        query = self.data_families.get(query_path, {}).get(query_name)  # type: ignore
        if not query:
            raise LookupError(f"Could not find {query_name} in path {query_path}.")
        return query

    def find_query(self, query_name: str) -> Set[Optional[QuerySource]]:
        """
        Return set of queries with name `query_name`.

        Parameters
        ----------
        query_name: str
            Name of the query

        Returns
        -------
        Set[QuerySource]
            Set (distinct) queries matching name.

        """
        return {
            query_dict.get(query_name)
            for family, query_dict in self.data_families.items()
            if query_name in query_dict
        }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data obfuscation functions."""
import hashlib
import re
import uuid
import warnings
from functools import lru_cache
import pkgutil
from typing import Any, Callable, Dict, List, Mapping, Union, Optional, Tuple

import numpy as np
import pandas as pd
import yaml

OBFUS_COL_MAP: Dict[str, str] = {}
_MAP_FILE = "resources/obfuscation_cols.yaml"
_obfus_map_file = pkgutil.get_data("msticpy", _MAP_FILE)
if not _obfus_map_file:
    warnings.warn(f"Could not find obfuscation column map {_MAP_FILE}")
else:
    _obfus_dicts = yaml.safe_load(_obfus_map_file)
    for data_col_map in _obfus_dicts.values():
        OBFUS_COL_MAP.update(data_col_map)

# SHA256 instance for hashing
sha_func = hashlib.sha256()

# Create a random map for shuffling IP address components
rng = np.random.default_rng()
ip_list = [str(n) for n in np.arange(256)]
rand_list = ip_list.copy()
rng.shuffle(rand_list)
ip_map = dict(zip(ip_list, rand_list))


def hash_string(input_str: str) -> str:
    """
    Hash a simple string.

    Parameters
    ----------
    input_str : str
        The input string

    Returns
    -------
    str
        The obfuscated output string

    """
    if not input_str:
        return input_str
    if not isinstance(input_str, str):
        input_str = str(input_str)
    str_dig = hashlib.sha256(bytes(input_str, "utf-8")).hexdigest()
    factor = int(len(input_str) / len(str_dig))
    out_str = (str_dig * max(factor, 1))[: len(input_str)]
    if not input_str.isnumeric():
        return "".join([chr(int(c) + 105) if c.isdigit() else c for c in out_str])
    # Convert entirely numeric strings to numbers
    return "".join([c if c.isdigit() else str(ord(c) - 97) for c in out_str])


@lru_cache(maxsize=1024)
def hash_item(input_item: str, delim: str = None) -> str:
    """
    Hash a simple string.

    Parameters
    ----------
    input_item : str
        The input string
    delim: str, optional
        A string of delimiters to use to split the input string
        prior to hashing.

    Returns
    -------
    str
        The obfuscated output string

    """
    if not input_item or not isinstance(input_item, str):
        return input_item
    if not delim:
        return hash_string(input_item)
    if len(delim) == 1:
        return delim.join([hash_string(elem) for elem in input_item.split(delim)])

    out_str = input_item
    delim_char = delim[0]
    delim_rest = delim[1:]
    for str_item in out_str.split(delim_char):
        out_str = out_str.replace(str_item, hash_item(str_item, delim_rest))
    return out_str


@lru_cache(maxsize=1024)
def _hash_ip_item(ip_addr: str) -> str:
    """
    Hash IP address.

    Parameters
    ----------
    ip_addr : str
        IP address string

    Returns
    -------
    str
        Hashed IP Address.

    """
    if not ip_addr or not isinstance(ip_addr, str):
        return ip_addr
    if "." in ip_addr:
        return ".".join([ip_map.get(byte, "1") for byte in ip_addr.split(".")])
    if ":" in ip_addr:
        ip_out = []
        for part in ip_addr.split(":"):
            enc = hashlib.sha256(bytes(part, "utf-8")).hexdigest()[: len(part)]
            ip_out.append(enc)
        return ":".join(ip_out)
    return hashlib.sha256(bytes(ip_addr, "utf-8")).hexdigest()[: len(ip_addr)]


def hash_ip(input_item: Union[List[str], str]) -> Union[List[str], str]:
    """
    Hash IP address or list of IP addresses.

    Parameters
    ----------
    input_item : Union[List[str], str]
        List of IP addresses or single IP address.

    Returns
    -------
    Union[List[str], str]
        List of hashed addresses or single address.
        (depending on input)

    """
    if not input_item:
        return input_item
    if isinstance(input_item, list):
        return [_hash_ip_item(elem) for elem in input_item]
    return _hash_ip_item(input_item)


def hash_list(item_list: List[str]) -> List[str]:
    """
    Hash list of strings.

    Parameters
    ----------
    item_list : List[str]
        Input list

    Returns
    -------
    List[str]
        Hashed list

    """
    out_list = []
    for val in item_list:
        if isinstance(val, dict):
            hash_val = hash_dict(val)
        elif isinstance(val, list):
            hash_val = hash_list(val)
        else:
            hash_val = hash_string(val)
        out_list.append(hash_val)
    return out_list


def hash_dict(
    item_dict: Dict[str, Union[Dict[str, Any], List[Any], str]]
) -> Dict[str, Any]:
    """
    Hash dictionary values.

    Parameters
    ----------
    item_dict : Dict[str, Union[Dict[str, Any], List[Any], str]]
        Input item can be a Dict of strings, lists or other
        dictionaries.

    Returns
    -------
    Dict[str, Any]
        Dictionary with hashed values.

    """
    out_dict = {}
    for key, val in item_dict.items():
        if isinstance(val, dict):
            hash_val = hash_dict(val)
        elif isinstance(val, list):
            hash_val = hash_list(val)  # type: ignore
        else:
            hash_val = hash_string(val)  # type: ignore
        out_dict[key] = hash_val
    return out_dict


WK_SID_PATTERN = re.compile(r"S(-\d+){3}$")
SID_PATTERN = re.compile(r"(S(?:-\d+){3})((?:-\d+){3})(-\d+)$")


@lru_cache(maxsize=1024)
def hash_sid(sid: str) -> str:
    """
    Hash a SID preserving well-known SIDs and the RID.

    Parameters
    ----------
    sid : str
        SID string

    Returns
    -------
    str
        Hashed SID

    """
    if re.match(WK_SID_PATTERN, sid):
        return sid
    usr_sid = re.match(SID_PATTERN, sid)
    if usr_sid:
        return (
            f"{usr_sid.groups()[0]}{hash_item(usr_sid.groups()[1], delim='-')}"
            + f"{usr_sid.groups()[2]}"
        )
    return sid


def _guid_replacer() -> Callable[[str], str]:
    """
    Closure for replace_guid.

    Returns
    -------
    Callable[[str], str]
        replace_guid function

    """
    guid_map: Dict[str, str] = {}

    def _replace_guid(guid: str) -> str:
        """
        Replace GUID/UUID with mapped random UUID.

        Parameters
        ----------
        guid : str
            Input UUID.

        Returns
        -------
        str
            Mapped UUID

        """
        if not guid or not isinstance(guid, str):
            return guid

        if guid in guid_map:
            return guid_map[guid]
        new_guid = str(uuid.uuid4())
        guid_map[guid] = new_guid
        return new_guid

    return _replace_guid


replace_guid = _guid_replacer()


# DataFrame obfuscation functions

# Map codes to functions
MAP_FUNCS: Dict[str, Union[str, Callable]] = {
    "uuid": replace_guid,
    "ip": hash_ip,
    "str": hash_string,
    "dict": hash_dict,
    "list": hash_list,
    "sid": hash_sid,
    "null": "null",
}


def obfuscate_df(
    data: pd.DataFrame, column_map: Mapping[str, Any] = None, use_default: bool = True
) -> pd.DataFrame:
    """
    Obfuscate columns of a DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        Input dataframe
    column_map : Mapping[str, Any], optional
        Custom column mapping, by default None
    use_default: bool
        If True use the built-in map (adding any custom
        mappings to this dictionary)

    Returns
    -------
    pd.DataFrame
        Obfuscated dataframe.

    """
    col_map = OBFUS_COL_MAP.copy() if use_default else {}
    if column_map is not None:
        col_map.update(column_map)

    out_df = data.copy()
    print("obfuscating columns:")
    for col_name in data.columns:
        if col_name not in col_map:
            continue
        col_type = col_map.get(col_name, "str")
        print(col_name, end=", ")
        map_func = MAP_FUNCS.get(col_type)
        try:
            if map_func == "null":
                data[col_name] = None
            elif map_func is not None and callable(map_func):
                out_df[col_name] = out_df.apply(
                    lambda x, col=col_name, func=map_func: func(x[col]), axis=1
                )
            else:
                out_df[col_name] = out_df.apply(
                    lambda x, col=col_name, c_type=col_type: hash_item(x[col], c_type),
                    axis=1,
                )
        except Exception as err:
            print(col_name, str(err))
            raise

    print("\ndone")
    return out_df


def check_obfuscation(
    data: pd.DataFrame, orig_data: pd.DataFrame, index: int = 0, silent=True
) -> Optional[Tuple[List[str], List[str]]]:
    """
    Check the obfuscation results for a row.

    Parameters
    ----------
    data : pd.DataFrame
        Obfuscated DataFrame
    orig_data : pd.DataFrame
        Original DataFrame
    index : int, optional
        The row to check, by default 0
    silent: bool
        If False the function returns no output and
        returns lists of changed and unchanged columns.
        By default, True

    Returns
    -------
    Optional[Tuple[List[str], List[str]]] :
        If silent is True returns a tuple of unchanged, changed
        items. If False, returns None.

    """
    unchanged = []
    obfuscated = []
    for col in sorted(data.columns):
        if data.iloc[index][col] == orig_data.iloc[index][col]:
            unchanged.append(f"{col}: {data.iloc[index][col]}")
        else:
            obfuscated.append(
                f"{col}:   {orig_data.iloc[index][col]} ----> {data.iloc[index][col]}"
            )
    if not silent:
        print("===== Start Check ====")
        print("Unchanged columns:")
        print("------------------")
        print("\n".join(unchanged))
        print("\nObfuscated columns:")
        print("--------------------")
        print("\n".join(obfuscated))
        print("====== End Check =====")
        return None

    return unchanged, obfuscated


@pd.api.extensions.register_dataframe_accessor("mp_obf")
class ObfuscationAccessor:
    """Base64 Unpack pandas extension."""

    def __init__(self, pandas_obj):
        """Initialize the extension."""
        self._df = pandas_obj

    def obfuscate(
        self, column_map: Mapping[str, Any] = None, use_default: bool = True
    ) -> pd.DataFrame:
        """
        Obfuscate the data in columns of a pandas dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            dataframe containing column to obfuscate
        column_map : Mapping[str, Any], optional
            Custom column mapping, by default None
        use_default: bool
            If True use the built-in map (adding any custom
            mappings to this dictionary)

        Returns
        -------
        pd.DataFrame
            Obfuscated dataframe

        """
        return obfuscate_df(
            data=self._df, column_map=column_map, use_default=use_default
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query hierarchy attribute class."""
from functools import partial
from typing import Any

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


class QueryContainer:
    """Empty class used to create hierarchical attributes."""

    def __len__(self):
        """Return number of items in the attribute collection."""
        return len(self.__dict__)

    def __iter__(self):
        """Return iterator over the attributes."""
        return iter(self.__dict__.items())

    def __getattr__(self, name):
        """Print usable error message if attribute not found."""
        if "." in name:
            try:
                attr = _get_dot_attrib(self, name)
            except KeyError:
                pass
            else:
                return attr
        raise AttributeError(
            f"Query attribute {name} not found.",
            "Use QueryProvider.list_queries() to see available queries.",
        )

    def __repr__(self):
        """Return list of attributes."""
        repr_list = []
        for name, obj in self.__dict__.items():
            if isinstance(obj, QueryContainer):
                repr_list.append(f"{name} (container)")
            elif isinstance(obj, partial):
                repr_list.append(f"{name} (query)")
        return "\n".join(repr_list)

    def __call__(self, *args, **kwargs):
        """Return list of attributes or help."""
        if args or kwargs:
            print(f"{self.__name__} is a container, not a query.")
            print("Items in this container:")
        print(repr(self))


def _get_dot_attrib(obj, elem_path: str) -> Any:
    """Return attribute at dotted path."""
    path_elems = elem_path.split(".")
    cur_node = obj
    for elem in path_elems:
        cur_node = getattr(cur_node, elem, None)
        if cur_node is None:
            raise KeyError(f"{elem} value of {elem_path} is not a valid path")
    return cur_node



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for TILookup classes.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import sys  # noqa
import warnings
from collections import ChainMap
from inspect import isclass
from typing import Dict, Iterable, List, Mapping, Optional, Tuple, Union

import attr
import pandas as pd

from .._version import VERSION
from ..common.exceptions import MsticpyConfigException, MsticpyUserConfigError
from ..common.provider_settings import get_provider_settings, reload_settings
from ..common.utility import export
from ..nbtools.ti_browser import browse_results
from . import tiproviders

# used in dynamic instantiation of providers
# pylint: disable=unused-wildcard-import, wildcard-import
from .tiproviders import *  # noqa:F401, F403
from .tiproviders.ti_provider_base import LookupResult, TILookupStatus, TIProvider

__version__ = VERSION
__author__ = "Ian Hellen"


_NO_PROVIDERS_MSSG = """
No TI Providers are loaded - please check that
you have correctly configured your msticpyconfig.yaml settings.
For more information see
https://msticpy.readthedocs.io/en/latest/TIProviders.html
"""


@export
class TILookup:
    """Threat Intel observable lookup from providers."""

    def __init__(
        self,
        primary_providers: Optional[List[TIProvider]] = None,
        secondary_providers: Optional[List[TIProvider]] = None,
        providers: Optional[List[str]] = None,
    ):
        """
        Initialize TILookup instance.

        Parameters
        ----------
        primary_providers : Optional[List[TIProvider]], optional
            Primary TI Providers, by default None
        secondary_providers : Optional[List[TIProvider]], optional
            Secondary TI Providers, by default None
        providers: Optional[List[str]], optional
            List of provider names to load, by default all available
            providers are loaded. To see the list of available providers
            call `TILookup.list_available_providers()`.
            Note: if primary_provides or secondary_providers is specified
            This will override the providers list.

        """
        self._providers: Dict[str, TIProvider] = {}
        self._secondary_providers: Dict[str, TIProvider] = {}
        self._providers_to_load = providers

        if primary_providers:
            for prov in primary_providers:
                self.add_provider(prov, primary=True)
        if secondary_providers:
            for prov in secondary_providers:
                self.add_provider(prov, primary=False)
        if not (primary_providers or secondary_providers):
            self._load_providers()

        self._all_providers = ChainMap(self._secondary_providers, self._providers)

    @property
    def loaded_providers(self) -> Dict[str, TIProvider]:
        """
        Return dictionary of loaded providers.

        Returns
        -------
        Dict[str, TIProvider]
            [description]

        """
        return self._all_providers  # type: ignore

    @property
    def provider_status(self) -> Iterable[str]:
        """
        Return loaded provider status.

        Returns
        -------
        Iterable[str]
            List of providers and descriptions.

        """
        prim = [
            f"{prov_name} - {prov.description} (primary)"
            for prov_name, prov in self._providers.items()
        ]
        sec = [
            f"{prov_name} - {prov.description} (secondary)"
            for prov_name, prov in self._secondary_providers.items()
        ]
        return prim + sec

    @property
    def configured_providers(self) -> List[str]:
        """
        Return a list of avaliable providers that have configuration details present.

        Returns
        -------
        List[str]
            List of TI Provider classes.

        """
        prim_conf = list(self._providers.keys())
        sec_conf = list(self._secondary_providers.keys())

        return prim_conf + sec_conf

    @property
    def available_providers(self) -> List[str]:
        """
        Return a list of builtin providers.

        Returns
        -------
        List[str]
            List of TI Provider classes.

        """
        return self._get_available_providers()

    @classmethod
    def _get_available_providers(cls):
        providers = []
        for provider_name in dir(tiproviders):
            provider_class = getattr(tiproviders, provider_name, None)
            if not (provider_class and isclass(provider_class)):
                continue
            # if it is a class - we only want to show concrete classes
            # that are sub-classes of TIProvider
            if issubclass(provider_class, tiproviders.TIProvider) and not bool(
                getattr(provider_class, "__abstractmethods__", False)
            ):
                providers.append(provider_class.__name__)
        return providers

    @classmethod
    def list_available_providers(
        cls, show_query_types=False, as_list: bool = False
    ) -> Optional[List[str]]:  # type: ignore
        """
        Print a list of builtin providers with optional usage.

        Parameters
        ----------
        show_query_types : bool, optional
            Show query types supported by providers, by default False
        as_list : bool, optional
            Return list of providers instead of printing to stdout.
            Note: if you specify `show_query_types` this will be printed
            irrespective of this parameter setting.

        Returns
        -------
        Optional[List[str]]
            A list of provider names (if `return_list=True`)

        """
        providers = []
        for provider_name in cls._get_available_providers():
            provider_class = getattr(tiproviders, provider_name, None)
            if not as_list:
                print(provider_name)
            providers.append(provider_name)
            if show_query_types:
                provider_class.usage()
        if as_list:
            return providers
        return None

    def provider_usage(self):
        """Print usage of loaded providers."""
        print("Primary providers")
        print("-----------------")
        if self._providers:
            for prov_name, prov in self._providers.items():
                print(f"\nProvider class: {prov_name}")
                prov.usage()
        else:
            print("none")
        print("\nSecondary providers")
        print("-------------------")
        if self._secondary_providers:
            for prov_name, prov in self._secondary_providers.items():
                print(f"\nProvider class: {prov_name}")
                prov.usage()
        else:
            print("none")

    @classmethod
    def reload_provider_settings(cls):
        """Reload provider settings from config."""
        reload_settings()
        print(
            "Settings reloaded. Use reload_providers to update settings",
            "for loaded providers.",
        )

    def reload_providers(self):
        """
        Reload providers based on currrent settings in config.

        Parameters
        ----------
        clear_keyring : bool, optional
            Clears any secrets cached in keyring, by default False

        """
        self.reload_provider_settings()
        self._load_providers()

    def _load_providers(self):
        """Load provider classes based on config."""
        prov_settings = get_provider_settings()

        for provider_entry, settings in prov_settings.items():
            # Allow overriding provider name to use another class
            provider_name = settings.provider if settings.provider else provider_entry
            if self._providers_to_load and provider_name not in self._providers_to_load:
                continue
            provider_class: TIProvider = getattr(
                sys.modules[__name__], provider_name, None
            )
            if not provider_class:
                warnings.warn(
                    f"Could not find provider class for {provider_name} "
                    + f"in config section {provider_entry}"
                )
                continue

            # instantiate class sending args from settings to init
            try:
                provider_instance = provider_class(**(settings.args))
            except MsticpyConfigException as mp_ex:
                # If the TI Provider didn't load, raise an exception
                raise MsticpyUserConfigError(
                    f"Could not load TI Provider {provider_name}",
                    *mp_ex.args,
                    "To avoid loading this provider please use the 'providers' parameter"
                    + " to TILookup() to specify which providers to load.",
                    title="TIProvider configuration error",
                    help_uri="https://msticpy.readthedocs.io/en/latest/data_acquisition/"
                    + "TIProviders.html#configuration-file",
                ) from mp_ex

            # set the description from settings, if one is provided, otherwise
            # use class docstring.
            provider_instance.description = (
                settings.description
                if settings.description
                else provider_instance.__doc__
            )
            self.add_provider(
                provider=provider_instance, name=provider_name, primary=settings.primary
            )

    def add_provider(
        self, provider: TIProvider, name: str = None, primary: bool = True
    ):
        """
        Add a TI provider to the current collection.

        Parameters
        ----------
        provider : TIProvider
            Provider instance
        name : str, optional
            The name to use for the provider (overrides the class name
            of `provider`)
        primary : bool, optional
            "primary" or "secondary" if False, by default "primary"

        """
        if not name:
            name = provider.__class__.__name__
        if primary:
            self._providers[name] = provider
        else:
            self._secondary_providers[name] = provider

    # pylint: disable=too-many-arguments
    def lookup_ioc(
        self,
        observable: str = None,
        ioc_type: str = None,
        ioc_query_type: str = None,
        providers: List[str] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> Tuple[bool, List[Tuple[str, LookupResult]]]:
        """
        Lookup single IoC in active providers.

        Parameters
        ----------
        observable : str
            IoC observable
            (`ioc` is also an alias for observable)
        ioc_type : str, optional
            One of IoCExtract.IoCType, by default None
            If none, the IoC type will be inferred
        ioc_query_type: str, optional
            The ioc query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        Tuple[bool, List[Tuple[str, LookupResult]]]
            The result returned as a tuple(bool, list):
            bool indicates whether a TI record was found in any provider
            list has an entry for each provider result

        """
        if not observable and "ioc" in kwargs:
            observable = kwargs["ioc"]
        if not observable:
            raise ValueError("observable or ioc parameter must be supplied.")

        result_list: List[Tuple[str, LookupResult]] = []
        selected_providers = self._select_providers(providers, prov_scope)
        if not selected_providers:
            raise RuntimeError(_NO_PROVIDERS_MSSG)

        ioc_type = ioc_type if ioc_type else TIProvider.resolve_ioc_type(observable)
        for prov_name, provider in selected_providers.items():
            provider_result: LookupResult = provider.lookup_ioc(
                ioc=observable, ioc_type=ioc_type, query_type=ioc_query_type, **kwargs
            )
            result_list.append((prov_name, provider_result))
        overall_result = any(res.result for _, res in result_list)
        return overall_result, result_list

    def lookup_iocs(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        obs_col: str = None,
        ioc_type_col: str = None,
        ioc_query_type: str = None,
        providers: List[str] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a collection of IoCs.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Mapping (e.g. a dict) of [observable, IoCType]
            3. Iterable of observables - IoCTypes will be inferred
        obs_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        ioc_query_type: str, optional
            The ioc query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        result_list: List[pd.DataFrame] = []
        selected_providers = self._select_providers(providers, prov_scope)
        if not selected_providers:
            raise RuntimeError(_NO_PROVIDERS_MSSG)

        for prov_name, provider in selected_providers.items():
            provider_result = provider.lookup_iocs(
                data=data,
                obs_col=obs_col,
                ioc_type_col=ioc_type_col,
                query_type=ioc_query_type,
                **kwargs,
            )
            if provider_result is None or provider_result.empty:
                continue
            if not kwargs.get("show_not_supported", False):
                provider_result = provider_result[
                    provider_result["Status"] != TILookupStatus.not_supported.value
                ]
            if not kwargs.get("show_bad_ioc", False):
                provider_result = provider_result[
                    provider_result["Status"] != TILookupStatus.bad_format.value
                ]
            provider_result["Provider"] = prov_name
            result_list.append(provider_result)

        if not result_list:
            print("No IoC matches")
        return pd.concat(result_list, sort=False)

    @staticmethod
    def result_to_df(
        ioc_lookup: Tuple[bool, List[Tuple[str, LookupResult]]]
    ) -> pd.DataFrame:
        """
        Return DataFrame representation of IoC Lookup response.

        Parameters
        ----------
        ioc_lookup : Tuple[bool, List[Tuple[str, LookupResult]]]
            Output from `lookup_ioc`

        Returns
        -------
        pd.DataFrame
            The response as a DataFrame with a row for each
            provider response.

        """
        return (
            pd.DataFrame(
                {
                    r_item[0]: pd.Series(attr.asdict(r_item[1]))
                    for r_item in ioc_lookup[1]
                }
            )
            .T.rename(columns=LookupResult.column_map())
            .drop("SafeIoc", axis=1)
        )

    def _select_providers(
        self, providers: List[str] = None, prov_scope: str = "primary"
    ) -> Dict[str, TIProvider]:
        """
        Return required subset of providers.

        Parameters
        ----------
        providers : List[str], optional
            Explicit list of provider names, by default None
        prov_scope : str, optional
            Provider scope, by default "primary"
            Other values are "all" and "secondary"

        Returns
        -------
        Dict[str, TIProvider]
            Dictionary of provider names and instances.

        """
        if providers:
            selected_providers = {
                prov_name: prov
                for prov_name, prov in self._all_providers.items()
                if prov_name in providers
            }
        else:
            if prov_scope == "all":
                selected_providers = self._all_providers  # type: ignore
            elif prov_scope == "primary":
                selected_providers = self._providers
            else:
                selected_providers = self._secondary_providers
        return selected_providers

    @staticmethod
    def browse_results(
        data: pd.DataFrame, severities: Optional[List[str]] = None, **kwargs
    ):
        """
        Return TI Results list browser.

        Parameters
        ----------
        data : pd.DataFrame
            TI Results data from TIProviders
        severities : Optional[List[str]], optional
            A list of the severity classes to show.
            By default these are ['warning', 'high'].
            Pass ['information', 'warning', 'high'] to see all
            results.

        Other Parameters
        ----------------
        kwargs :
            passed to SelectItem constuctor.

        Returns
        -------
        SelectItem
            SelectItem browser for TI Data.

        """
        return browse_results(data=data, severities=severities, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Helper module for computations when modelling sessions.

In particular, this module is for when each session is a list of the
Cmd datatype with the params attribute set to a dictionary of
accompanying params and values.
"""

from collections import defaultdict
from typing import Tuple, List, Union, DefaultDict
import copy

import numpy as np

from ..utils.data_structures import StateMatrix, Cmd
from ..utils.laplace_smooth import (
    laplace_smooth_cmd_counts,
    laplace_smooth_param_counts,
    laplace_smooth_value_counts,
)
from ....common.exceptions import MsticpyException


# pylint: disable=too-many-locals, too-many-branches
def compute_counts(  # noqa MC0001  # nosec
    sessions: List[List[Cmd]], start_token: str, end_token: str
) -> Tuple[
    DefaultDict[str, int],
    DefaultDict[str, DefaultDict[str, int]],
    DefaultDict[str, int],
    DefaultDict[str, DefaultDict[str, int]],
    DefaultDict[str, int],
    DefaultDict[str, DefaultDict[str, int]],
]:
    """
    Compute the training counts for the sessions.

    In particular, computes counts of individual commands and of sequences
    of two commands. It also computes the counts of individual params as
    well as counts of params conditional on the command. It also computes
    the counts of individual values as well as counts of values conditional
    on the param.

    Parameters
    ----------
    sessions: List[List[Cmd]]
        each session is a list of the Cmd datatype.
        Where the Cmd datatype has a name attribute (command name) and
        a params attribute (dict with the params and values associated
        with the command)
        an example session:
            [
                Cmd(
                    name='Set-User',
                    params={'Identity': 'blahblah', 'Force': 'true'}
                ),
                Cmd(
                    name='Set-Mailbox',
                    params={'Identity': 'blahblah', 'AuditEnabled': 'false'}
                )
            ]
    start_token: str
        dummy command to signify the start of a session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of a session (e.g. "##END##")

    Returns
    -------
    tuple of counts:
        individual command counts,
        sequence command (length 2) counts,
        individual param counts,
        param conditional on command counts
        individual value counts,
        value conditional on param counts

    """
    seq1_counts: DefaultDict[str, int] = defaultdict(lambda: 0)
    seq2_counts: DefaultDict[str, DefaultDict[str, int]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )

    param_counts: DefaultDict[str, int] = defaultdict(lambda: 0)
    cmd_param_counts: DefaultDict[str, DefaultDict[str, int]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )

    value_counts: DefaultDict[str, int] = defaultdict(lambda: 0)
    param_value_counts: DefaultDict[str, DefaultDict[str, int]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )

    for session in sessions:
        prev = start_token
        seq1_counts[prev] += 1
        for cmd in session:
            seq1_counts[cmd.name] += 1
            seq2_counts[prev][cmd.name] += 1
            prev = cmd.name
            pars = cmd.params
            if isinstance(pars, set):
                pars = dict.fromkeys(pars)
            for par, val in pars.items():
                param_counts[par] += 1
                value_counts[val] += 1
                cmd_param_counts[cmd.name][par] += 1
                param_value_counts[par][val] += 1
        seq2_counts[prev][end_token] += 1
        seq1_counts[end_token] += 1

    return (
        seq1_counts,
        seq2_counts,
        param_counts,
        cmd_param_counts,
        value_counts,
        param_value_counts,
    )


# pylint: disable=too-many-arguments
def laplace_smooth_counts(
    seq1_counts: DefaultDict[str, int],
    seq2_counts: DefaultDict[str, DefaultDict[str, int]],
    param_counts: DefaultDict[str, int],
    cmd_param_counts: DefaultDict[str, DefaultDict[str, int]],
    value_counts: DefaultDict[str, int],
    param_value_counts: DefaultDict[str, DefaultDict[str, int]],
    start_token: str,
    end_token: str,
    unk_token: str,
) -> Tuple[
    StateMatrix, StateMatrix, StateMatrix, StateMatrix, StateMatrix, StateMatrix
]:
    """
    Laplace smoothing is applied to the counts.

    We do this by adding 1 to each of the counts.
    This is so we shift some of the probability mass from the very probable
    commands/params/values to the unseen and very unlikely commands/params/values.
    The `unk_token` means we can handle unseen commands, params, values, sequences of
    commands.

    Parameters
    ----------
    seq1_counts: DefaultDict[str, int]
        individual command counts
    seq2_counts: DefaultDict[str, DefaultDict[str, int]]
        sequence command (length 2) counts
    param_counts: DefaultDict[str, int]
        individual param counts
    cmd_param_counts: DefaultDict[str, DefaultDict[str, int]]
        param conditional on command counts
    value_counts: DefaultDict[str, int]
        individual value counts
    param_value_counts: DefaultDict[str, DefaultDict[str, int]]
        value conditional on param counts
    start_token: str
        dummy command to signify the start of a session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of a session (e.g. "##END##")
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    tuple of StateMatrix counts:
        individual command counts,
        sequence command (length 2) counts,
        individual param counts,
        param conditional on command counts
        individual value counts,
        value conditional on param counts

    """
    cmds: List[str] = list(seq1_counts.keys()) + [unk_token]

    # apply laplace smoothing to the cmds
    seq1_counts_ls, seq2_counts_ls = laplace_smooth_cmd_counts(
        seq1_counts=copy.deepcopy(seq1_counts),
        seq2_counts=copy.deepcopy(seq2_counts),
        start_token=start_token,
        end_token=end_token,
        unk_token=unk_token,
    )

    params: List[str] = list(param_counts.keys()) + [unk_token]

    # apply laplace smoothing to the params
    param_counts_ls, cmd_param_counts_ls = laplace_smooth_param_counts(
        cmds=cmds,
        param_counts=copy.deepcopy(param_counts),
        cmd_param_counts=copy.deepcopy(cmd_param_counts),
        unk_token=unk_token,
    )

    # apply laplace smoothing for the values
    value_counts_ls, param_value_counts_ls = laplace_smooth_value_counts(
        params=params,
        value_counts=copy.deepcopy(value_counts),
        param_value_counts=copy.deepcopy(param_value_counts),
        unk_token=unk_token,
    )

    seq1_counts_sm = StateMatrix(states=seq1_counts_ls, unk_token=unk_token)
    seq2_counts_sm = StateMatrix(states=seq2_counts_ls, unk_token=unk_token)
    param_counts_sm = StateMatrix(states=param_counts_ls, unk_token=unk_token)
    cmd_param_counts_sm = StateMatrix(states=cmd_param_counts_ls, unk_token=unk_token)
    value_counts_sm = StateMatrix(states=value_counts_ls, unk_token=unk_token)
    param_value_counts_sm = StateMatrix(
        states=param_value_counts_ls, unk_token=unk_token
    )

    return (
        seq1_counts_sm,
        seq2_counts_sm,
        param_counts_sm,
        cmd_param_counts_sm,
        value_counts_sm,
        param_value_counts_sm,
    )


def get_params_to_model_values(
    param_counts: Union[StateMatrix, dict], param_value_counts: Union[StateMatrix, dict]
) -> set:
    """
    Determine using heuristics which params take categoricals vs arbitrary strings.

    This function helps us decide which params we should model
    the values of later on.

    Parameters
    ----------
    param_counts: Union[StateMatrix, dict]
        counts of each of the individual params
    param_value_counts: Union[StateMatrix, dict]
        counts of each value conditional on the params

    Returns
    -------
    set of params which have been determined to be categorical

    """
    param_stats = [
        (param, len(vals), param_counts[param], 100 * len(vals) / param_counts[param])
        for param, vals in param_value_counts.items()
    ]

    modellable_params = [
        param[0]
        for param in param_stats
        if param[1] <= 20 <= param[2] and param[3] <= 10
    ]

    return set(modellable_params)


# pylint: disable=too-many-arguments, too-many-branches
def compute_prob_setofparams_given_cmd(
    cmd: str,
    params_with_vals: Union[dict, set],
    param_cond_cmd_probs: Union[StateMatrix, dict],
    value_cond_param_probs: Union[StateMatrix, dict],
    modellable_params: Union[set, list],
    use_geo_mean: bool = True,
) -> float:
    """
    Compute probability of a set of params + values given the cmd.

    Parameters
    ----------
    cmd: str
        name of command (e.g. for Exchange powershell commands: "Set-Mailbox")
    params_with_vals: Union[dict, set]
        dict of accompanying params and values for the cmd
        e.g for Exchange powershell commands:
            {'Identity': 'an_identity' , 'ForwardingEmailAddress': 'email@email.com'}
            If params is set to be a set, then an artificial dictionary will
            be created with the set as the keys and Nones for the values.
    param_cond_cmd_probs: Union[StateMatrix, dict]
        computed probabilities of params conditional on the command
    value_cond_param_probs: Union[StateMatrix, dict]
        computed probabilities of values conditional on the param
    modellable_params: set
        set of params for which we will also include the probabilties
        of their values in the calculation of the likelihood
    use_geo_mean: bool
        if True, then the likelihood will be raised to the power of (1/K)
        where K is the number of distinct params which appeared for the
        given `cmd` across our training set + the number of
        values which we included in the modelling for this cmd.
        Note: some commands may have more params set in general compared
        with other commands. It can be useful to use the geo mean
        so that you can compare this probability across different
        commands with differing number of params.

    Returns
    -------
    computed probability

    """
    pars = params_with_vals.copy()
    if isinstance(pars, set):
        pars = dict.fromkeys(pars)
    if len(pars) == 0:
        return 1.0
    ref_cmd = param_cond_cmd_probs[cmd]
    lik: float = 1
    num = 0
    for param, prob in ref_cmd.items():
        if param in pars:
            lik *= prob
            if param in modellable_params:
                num += 1
                val = pars[param]
                lik *= value_cond_param_probs[param][val]
        else:
            lik *= 1 - prob
    if use_geo_mean:
        k = len(ref_cmd) + num
        if k > 0:
            lik = lik ** (1 / k)

    return lik


# pylint: disable=too-many-locals, too-many-arguments, too-many-branches
def compute_likelihood_window(
    window: List[Cmd],
    prior_probs: Union[StateMatrix, dict],
    trans_probs: Union[StateMatrix, dict],
    param_cond_cmd_probs: Union[StateMatrix, dict],
    value_cond_param_probs: Union[StateMatrix, dict],
    modellable_params: set,
    use_start_token: bool,
    use_end_token: bool,
    start_token: str = None,
    end_token: str = None,
) -> float:
    """
    Compute the likelihood of the input `window`.

    Parameters
    ----------
    window: List[Cmd]
        part or all of a session, where a session is a list the Cmd datatype
        an example session:
            [
                Cmd(name='Set-User', params={'Identity': 'blahblah', 'Force': 'true'}),
                Cmd(name='Set-Mailbox',
                    params={'Identity': 'blahblah', 'AuditEnabled': 'false'})
            ]
    prior_probs: Union[StateMatrix, dict]
        computed probabilities of individual commands
    trans_probs: Union[StateMatrix, dict]
        computed probabilities of sequences of commands (length 2)
    param_cond_cmd_probs: Union[StateMatrix, dict]
        computed probabilities of the params conditional on the commands
    value_cond_param_probs: Union[StateMatrix, dict]
        computed probabilities of the values conditional on the params
    modellable_params: set
        set of params for which we will also include the probabilties
        of their values in the calculation of the likelihood
    use_start_token: bool
        if set to True, the start_token will be prepended to the
        window before the likelihood calculation is done
    use_end_token: bool
        if set to True, the end_token will be appended to the window
        before the likelihood calculation is done
    start_token: str
        dummy command to signify the start of the session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of the session (e.g. "##END##")

    Returns
    -------
    likelihood of the window

    """
    if use_start_token:
        if start_token is None:
            raise MsticpyException(
                "start_token should not be None, when use_start_token is True"
            )
    if use_end_token:
        if end_token is None:
            raise MsticpyException(
                "end_token should not be None, when use_end_token is True"
            )

    w_len = len(window)
    if w_len == 0:
        return np.nan
    prob: float = 1

    cur_cmd = window[0].name
    params = window[0].params
    param_vals_prob = compute_prob_setofparams_given_cmd(
        cmd=cur_cmd,
        params_with_vals=params,
        param_cond_cmd_probs=param_cond_cmd_probs,
        value_cond_param_probs=value_cond_param_probs,
        modellable_params=modellable_params,
        use_geo_mean=True,
    )

    if use_start_token:
        prob *= trans_probs[start_token][cur_cmd] * param_vals_prob
    else:
        prob *= prior_probs[cur_cmd] * param_vals_prob

    for i in range(1, w_len):
        prev, cur = window[i - 1], window[i]
        prev_cmd, cur_cmd = prev.name, cur.name
        cur_par = cur.params
        prob *= trans_probs[prev_cmd][cur_cmd]
        param_vals_prob = compute_prob_setofparams_given_cmd(
            cmd=cur_cmd,
            params_with_vals=cur_par,
            param_cond_cmd_probs=param_cond_cmd_probs,
            value_cond_param_probs=value_cond_param_probs,
            modellable_params=modellable_params,
            use_geo_mean=True,
        )
        prob *= param_vals_prob

    if use_end_token:
        prob *= trans_probs[cur_cmd][end_token]

    return prob


# pylint: disable=too-many-locals, too-many-arguments
def compute_likelihood_windows_in_session(
    session: List[Cmd],
    prior_probs: Union[StateMatrix, dict],
    trans_probs: Union[StateMatrix, dict],
    param_cond_cmd_probs: Union[StateMatrix, dict],
    value_cond_param_probs: Union[StateMatrix, dict],
    modellable_params: set,
    window_len: int,
    use_start_end_tokens: bool,
    start_token: str = None,
    end_token: str = None,
    use_geo_mean: bool = False,
) -> List[float]:
    """
    Compute the likelihoods of a sliding window of `window_len` in the session.

    Parameters
    ----------
    session: List[Cmd]
        list of Cmd datatype
        an example session:
            [
                Cmd(
                    name='Set-User',
                    params={'Identity': 'blahblah', 'Force': 'true'}
                ),
                Cmd(
                    name='Set-Mailbox',
                    params={'Identity': 'blahblah', 'AuditEnabled': 'false'}
                )
            ]
    prior_probs: Union[StateMatrix, dict]
        computed probabilities of individual commands
    trans_probs: Union[StateMatrix, dict]
        computed probabilities of sequences of commands (length 2)
    param_cond_cmd_probs: Union[StateMatrix, dict]
        computed probabilities of the params conditional on the commands
    value_cond_param_probs: Union[StateMatrix, dict]
        computed probabilities of the values conditional on the params
    modellable_params: set
        set of params for which we will also include the probabilties
        of their values in the calculation of the likelihood
    window_len: int
        length of sliding window for likelihood calculations
    use_start_end_tokens: bool
        if True, then `start_token` and `end_token` will be prepended
        and appended to the session respectively before the calculations
        are done
    start_token: str
        dummy command to signify the start of the session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of the session (e.g. "##END##")
    use_geo_mean: bool
        if True, then each of the likelihoods of the sliding windows
        will be raised to the power of (1/`window_len`)

    Returns
    -------
    list of likelihoods

    """
    if use_start_end_tokens:
        if start_token is None or end_token is None:
            raise MsticpyException(
                "start_token and end_token should not be set to None when "
                "use_start_end_tokens is set to True"
            )

    likelihoods = []
    sess = session.copy()
    if use_start_end_tokens and end_token:
        sess += [Cmd(name=str(end_token), params=dict())]
    end = len(sess) - window_len
    for i in range(end + 1):
        window = sess[i : i + window_len]  # noqa E203
        if i == 0:
            use_start = use_start_end_tokens
        else:
            use_start = False
        lik = compute_likelihood_window(
            window=window,
            prior_probs=prior_probs,
            trans_probs=trans_probs,
            param_cond_cmd_probs=param_cond_cmd_probs,
            value_cond_param_probs=value_cond_param_probs,
            modellable_params=modellable_params,
            use_start_token=use_start,
            use_end_token=False,
            start_token=start_token,
            end_token=end_token,
        )
        if use_geo_mean:
            k = window_len
            lik = lik ** (1 / k)
        likelihoods.append(lik)

    return likelihoods


# pylint: disable=too-many-arguments
def rarest_window_session(
    session: List[Cmd],
    prior_probs: Union[StateMatrix, dict],
    trans_probs: Union[StateMatrix, dict],
    param_cond_cmd_probs: Union[StateMatrix, dict],
    value_cond_param_probs: Union[StateMatrix, dict],
    modellable_params: set,
    window_len: int,
    use_start_end_tokens: bool,
    start_token: str,
    end_token: str,
    use_geo_mean: bool = False,
) -> Tuple[List[Cmd], float]:
    """
    Find and compute likelihood of the rarest window of `window_len` in the session.

    Parameters
    ----------
    session: List[Cmd]
        list of Cmd datatype
        an example session:
            [
                Cmd(
                    name='Set-User',
                    params={'Identity': 'blahblah', 'Force': 'true'}
                ),
                Cmd(
                    name='Set-Mailbox',
                    params={'Identity': 'blahblah', 'AuditEnabled': 'false'}
                )
            ]
    prior_probs: Union[StateMatrix, dict]
        computed probabilities of individual commands
    trans_probs: Union[StateMatrix, dict]
        computed probabilities of sequences of commands (length 2)
    param_cond_cmd_probs: Union[StateMatrix, dict]
        computed probabilities of the params conditional on the commands
    value_cond_param_probs: Union[StateMatrix, dict]
        computed probabilities of the values conditional on the params
    modellable_params: set
        set of params for which we will also include the probabilties of
        their values in the calculation of the likelihood
    window_len: int
        length of sliding window for likelihood calculations
    use_start_end_tokens: bool
        if True, then `start_token` and `end_token` will be prepended
        and appended to the session respectively before the calculations are done
    start_token: str
        dummy command to signify the start of the session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of the session (e.g. "##END##")
    use_geo_mean: bool
        if True, then each of the likelihoods of the sliding windows
        will be raised to the power of (1/`window_len`)

    Returns
    -------
    Tuple:
        rarest window part of the session,
        likelihood of the rarest window

    """
    likelihoods = compute_likelihood_windows_in_session(
        session=session,
        prior_probs=prior_probs,
        trans_probs=trans_probs,
        param_cond_cmd_probs=param_cond_cmd_probs,
        value_cond_param_probs=value_cond_param_probs,
        modellable_params=modellable_params,
        window_len=window_len,
        use_start_end_tokens=use_start_end_tokens,
        start_token=start_token,
        end_token=end_token,
        use_geo_mean=use_geo_mean,
    )
    if len(likelihoods) == 0:
        return [], np.nan
    min_lik = min(likelihoods)
    ind = likelihoods.index(min_lik)
    return session[ind : ind + window_len], min_lik  # noqa E203



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MDATP OData Driver class."""
from typing import Union, Any
import pandas as pd

from .odata_driver import OData, QuerySource
from ...common.utility import export
from ..._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"


@export
class MDATPDriver(OData):
    """KqlDriver class to retreive date from MDATP."""

    def __init__(self, connection_str: str = None, **kwargs):
        """
        Instantiaite MDATPDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        """
        super().__init__()
        self.req_body = {
            "client_id": None,
            "client_secret": None,
            "grant_type": "client_credentials",
            "resource": "https://api.securitycenter.windows.com",
        }
        self.oauth_url = "https://login.windows.net/{tenantId}/oauth2/token"
        self.api_root = "https://api.securitycenter.windows.com/"
        self.api_ver = "api"

        if connection_str:
            self.current_connection = connection_str
            self.connect(connection_str)

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """
        del query_source, kwargs
        return self.query_with_results(
            query, body=True, api_end="/advancedqueries/run"
        )[0]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Auditd extractor.

Module to load and decode Linux audit logs. It collapses messages
sharing the same message ID into single events, decodes hex-encoded data
fields and performs some event-specific formatting and normalization
(e.g. for process start events it will re-assemble the process command
line arguments into a single string). This is still a work-in-progress.

"""
import codecs
import re
from datetime import datetime
from typing import Mapping, Any, Tuple, Dict, List, Optional, Set
import pandas as pd

from .process_tree_utils import build_process_tree
from .eventcluster import dbcluster_events, add_process_features


from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

# Constants
# Fields that we know are frequently encoded
_ENCODED_PARAMS: Dict[str, Set[str]] = {
    "EXECVE": {"a0", "a1", "a2", "a3", "arch"},
    "PROCTITLE": {"proctitle"},
    "USER_CMD": {"cmd"},
}

# USER_START message schema
_USER_START: Dict[str, Optional[str]] = {
    "pid": "int",
    "uid": "int",
    "auid": "int",
    "ses": "int",
    "msg": None,
    "acct": None,
    "exe": None,
    "hostname": None,
    "addr": None,
    "terminal": None,
    "res": None,
}

# Message types schema
_FIELD_DEFS: Dict[str, Dict[str, Optional[str]]] = {
    "SYSCALL": {
        "success": None,
        "ppid": "int",
        "pid": "int",
        "auid": "int",
        "uid": "int",
        "gid": "int",
        "euid": "int",
        "egid": "int",
        "ses": "int",
        "exe": None,
        "com": None,
    },
    "CWD": {"cwd": None},
    "PROCTITLE": {"proctitle": None},
    "LOGIN": {
        "pid": "int",
        "uid": "int",
        "tty": None,
        "old-ses": "int",
        "ses": "int",
        "res": None,
    },
    "EXECVE": {"argc": "int", "a0": None, "a1": None, "a2": None},
    "_USER_START": _USER_START,
    "USER_END": _USER_START,
    "CRED_DISP": _USER_START,
    "USER_ACCT": _USER_START,
    "CRED_ACQ": _USER_START,
    "USER_CMD": {
        "pid": "int",
        "uid": "int",
        "auid": "int",
        "ses": "int",
        "msg": None,
        "cmd": None,
        "terminal": None,
        "res": None,
    },
}


def unpack_auditd(audit_str: List[Dict[str, str]]) -> Mapping[str, Mapping[str, Any]]:
    """
    Unpack an Audit message and returns a dictionary of fields.

    Parameters
    ----------
    audit_str : str
        The auditd raw record

    Returns
    -------
    Mapping[str, Any]
        The extracted message fields and values

    """
    event_dict: Dict[str, Dict[str, Any]] = {}
    # The audit_str should be a list of dicts - '{EXECVE : {'p1': 'foo', p2: 'bar'...},
    #                                      PATH: {'a1': 'xyz',....}}

    for record in audit_str:
        # process a single message type, splitting into type name
        # and contents
        for rec_key, rec_val in record.items():
            rec_dict: Dict[str, Optional[str]] = {}
            # Get our field mapping for encoded params for this
            # mssg_type (rec_key)
            encoded_fields_map = _ENCODED_PARAMS.get(rec_key, None)
            for rec_item in rec_val:
                # for each mssg item, split into k/v pair
                rec_split = rec_item.split("=", maxsplit=1)
                if len(rec_split) == 1:
                    rec_dict[rec_split[0]] = None
                    continue
                if (
                    not encoded_fields_map
                    or rec_split[1].startswith('"')
                    or rec_split[0] not in encoded_fields_map
                ):
                    field_value = rec_split[1].strip('"')
                else:
                    try:
                        # Try to decode this from hex-string to text
                        # Mypy thinks codecs.decode returns a str so
                        # incorrectly issues a type warning - in this case it
                        # will return a bytes string.
                        field_value = codecs.decode(  # type: ignore
                            bytes(rec_split[1], "utf-8"), "hex"
                        ).decode("utf-8")
                    except ValueError:
                        field_value = rec_split[1]
                        print(rec_val)
                        print(
                            "ERR:",
                            rec_key,
                            rec_split[0],
                            rec_split[1],
                            type(rec_split[1]),
                        )
                rec_dict[rec_split[0]] = field_value
            event_dict[rec_key] = rec_dict

    return event_dict


def _extract_event(message_dict: Mapping[str, Any]) -> Tuple[str, Mapping[str, Any]]:
    """
    Assemble discrete messages sharing the same message Id into a single event.

    Parameters
    ----------
    message_dict : Mapping[str, Any]
        the input dictionary

    Returns
    -------
    Tuple[str, Mapping[str, Any]
        the assembled message type and contents

    """
    # Handle process executions specially
    if "SYSCALL" in message_dict and "EXECVE" in message_dict:
        proc_create_dict: Dict[str, Any] = {}
        for mssg_type in ["SYSCALL", "CWD", "EXECVE", "PROCTITLE"]:
            if mssg_type not in message_dict or mssg_type not in _FIELD_DEFS:
                continue
            _extract_mssg_value(mssg_type, message_dict, proc_create_dict)

            if mssg_type == "EXECVE":
                args = int(proc_create_dict.get("argc", 1))
                arg_strs = []
                for arg_idx in range(0, args):
                    arg_strs.append(proc_create_dict.get(f"a{arg_idx}", ""))

                proc_create_dict["cmdline"] = " ".join(arg_strs)
        return "SYSCALL_EXECVE", proc_create_dict

    event_dict: Dict[str, Any] = {}
    for mssg_type, _ in message_dict.items():
        if mssg_type in _FIELD_DEFS:
            _extract_mssg_value(mssg_type, message_dict, event_dict)
        else:
            # We don't check for duplicated keys here - if
            # there are multiple messages with the same key, the
            # last one will overwrite the previous value
            event_dict.update(message_dict[mssg_type])
    return list(message_dict.keys())[0], event_dict


def _extract_mssg_value(
    mssg_type: str,
    message_dict: Mapping[str, Mapping[str, Any]],
    event_dict: Dict[str, Any],
):
    """
    Extract field/value from the message dictionary.

    Parameters
    ----------
    mssg_type : str
        The Audit message type
    message_dict : Mapping[str, str]
        The input dictionary
    event_dict : Dict[str, Any]
        The output dictionary

    """
    # if the field requires conversion conv will specify the
    # target type - only int currently
    for fieldname, conv in _FIELD_DEFS[mssg_type].items():
        value = message_dict[mssg_type].get(fieldname, None)
        if not value:
            return
        if conv:
            if conv == "int":
                value = int(value)
                if value == 4294967295:
                    value = -1
        if fieldname in event_dict:
            event_dict[f"{fieldname}_{mssg_type}"] = value
        else:
            event_dict[fieldname] = value


def _move_cols_to_front(data: pd.DataFrame, column_count: int = 1) -> pd.DataFrame:
    """
    Move N columns from end to front of DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame
    column_count : int, optional
        The number of columns to move (the default is 1)

    Returns
    -------
    pd.DataFrame
        DataFrame with `column_count` columns shifted to front

    """
    return data[list(data.columns[-column_count:]) + list(data.columns[:-column_count])]


def extract_events_to_df(
    data: pd.DataFrame,
    input_column: str = "AuditdMessage",
    event_type: str = None,
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Extract auditd raw messages into a dataframe.

    Parameters
    ----------
    data : pd.DataFrame
        The input dataframe with raw auditd data in
        a single string column
    input_column : str, optional
        the input column name (the default is 'AuditdMessage')
    event_type : str, optional
        the event type, if None, defaults to all (the default is None)
    verbose : bool, optional
        Give feedback on stages of processing (the default is False)

    Returns
    -------
    pd.DataFrame
        The resultant DataFrame

    """
    if verbose:
        start_time = datetime.utcnow()
        print(f"Unpacking auditd messages for {len(data)} events...")

    # If the provided table has auditd messages as a string format and
    # extract key elements.
    if isinstance(data[input_column].head(1)[0], str):
        data["mssg_id"] = data.apply(
            lambda x: _extract_timestamp(x[input_column]), axis=1
        )
        data[input_column] = data.apply(
            lambda x: _parse_audit_message(x[input_column]), axis=1
        )

    # Our first pandas expression does most of the work - unpacking the
    # column contents, then extracting these into a two columns
    # EventType (the main auditd mssg type) and a dict of k/v values
    # EventData
    tmp_df = data.apply(
        lambda x: _extract_event(unpack_auditd(x[input_column])),
        axis=1,
        result_type="expand",
    ).rename(columns={0: "EventType", 1: "EventData"})
    # if only one type of event is requested
    if event_type:
        tmp_df = tmp_df.loc[tmp_df["EventType"] == event_type]
        if verbose:
            print(f"Event subset = {event_type} (events: {len(tmp_df)})")

    if verbose:
        print("Building output dataframe...")

    # We convert the EventData dict into a series,
    # then merge with:
    # First - the intermediate input DF to add back the EventType column
    # Second - the original input DF to add back metadata columns like Computer
    # Finally get rid of any empty columns
    tmp_df = (
        tmp_df.apply(lambda x: pd.Series(x.EventData), axis=1)
        .merge(tmp_df[["EventType"]], left_index=True, right_index=True)
        .merge(
            data.drop([input_column], axis=1),
            how="inner",
            left_index=True,
            right_index=True,
        )
        .dropna(axis=1, how="all")
    )

    if verbose:
        print("Fixing timestamps...")

    # extract real timestamp from mssg_id
    tmp_df["TimeStamp"] = tmp_df.apply(
        lambda x: datetime.utcfromtimestamp(float(x["mssg_id"].split(":")[0])), axis=1
    )
    if "TimeGenerated" in tmp_df:
        tmp_df = tmp_df.drop(["TimeGenerated"], axis=1)
    tmp_df = tmp_df.rename(columns={"TimeStamp": "TimeGenerated"}).pipe(
        _move_cols_to_front, column_count=5
    )
    if verbose:
        print(f"Complete. {len(tmp_df)} output rows", end=" ")
        delta = datetime.utcnow() - start_time
        print(f"time: {delta.seconds + delta.microseconds/1_000_000} sec")

    return tmp_df


def get_event_subset(data: pd.DataFrame, event_type: str) -> pd.DataFrame:
    """
    Return a subset of the events matching type event_type.

    Parameters
    ----------
    data : pd.DataFrame
        The input data
    event_type : str
        The event type to select

    Returns
    -------
    pd.DataFrame
        The subset of the data where
        data['EventType'] == event_type

    """
    return (
        data[data["EventType"] == event_type].dropna(axis=1, how="all").infer_objects()
    )


def read_from_file(
    filepath: str, event_type: str = None, verbose: bool = False, dummy_sep: str = "\t"
) -> pd.DataFrame:
    r"""
    Extract Audit events from a log file.

    Parameters
    ----------
    filepath : str
        path to the input file
    event_type : str, optional
        The type of event to extract if only a subset required.
        (the default is None, which processes all types)
    verbose : bool, optional
        If true more progress messages are output
        (the default is False)
    dummy_sep : str, optional
        Separator to use for reading the 'csv' file
        (default is tab - '\t')

    Returns
    -------
    pd.DataFrame
        The output DataFrame

    Notes
    -----
    The dummy_sep parameter should be a character that does not
    occur in an input line. This function uses pandas read_csv
    to read the audit lines into a single column. Using a separator
    that does appear in the input (e.g. space or comma) will cause
    data to be parsed into muliple columns and anything after the
    first separator in a line will be lost.

    """
    # read in the file using pd.read_csv()
    df_raw = pd.read_csv(
        filepath, sep=dummy_sep, names=["raw_data"], skip_blank_lines=True
    )

    # extract message ID into seperate column
    df_raw["mssg_id"] = df_raw.apply(
        lambda x: _extract_timestamp(x["raw_data"]), axis=1
    )
    # Pack message type and content into a dictionary:
    # {'mssg_type: ['item1=x, item2=y....]}
    df_raw["AuditdMessage"] = df_raw.apply(
        lambda x: _parse_audit_message(x["raw_data"]), axis=1
    )

    # Group the data by message id string and concatenate the message content
    # dictionaries in a list.
    df_grouped_cols = (
        df_raw.groupby(["mssg_id"]).agg({"AuditdMessage": list}).reset_index()
    )
    # pass this DataFrame to the event extractor.
    return extract_events_to_df(
        data=df_grouped_cols,
        input_column="AuditdMessage",
        event_type=event_type,
        verbose=verbose,
    )


def _parse_audit_message(audit_str: str) -> Dict[str, List[str]]:
    """
    Parse an auditd message string into Dict format required by unpack_auditd.

    Parameters
    ----------
    audit_str : str
        The Audit message

    Returns
    -------
    Dict[str, str]
        The extracted message values

    """
    audit_message = audit_str.rstrip().split(": ")
    audit_headers = audit_message[0]
    audit_hdr_match = re.match(r"type=([^\s]+)", audit_headers)
    if audit_hdr_match:
        audit_msg = {audit_hdr_match.group(1): audit_message[1].split(" ")}
        return audit_msg
    return {}  # type ignore


def _extract_timestamp(audit_str: str) -> str:
    """
    Parse an auditd message string and extract the message time.

    Parameters
    ----------
    audit_str : str
        The Audit message

    Returns
    -------
    str
        The extracted message time string

    """
    audit_message = audit_str.rstrip().split(": ")
    audit_headers = audit_message[0]
    audit_hdr_match = re.match(r".*msg=audit\(([^\)]+)\)", audit_headers)
    if audit_hdr_match:
        time_stamp = audit_hdr_match.group(1).split(":")[0]
        return time_stamp
    return ""


# pylint: disable=too-many-branches
def generate_process_tree(  # noqa: MC0001
    audit_data: pd.DataFrame, branch_depth: int = 4, processes: pd.DataFrame = None
) -> pd.DataFrame:
    """
    Generate process tree data from auditd logs.

    Parameters
    ----------
    audit_data : pd.DataFrame
        The Audit data containing process creation events
    branch_depth: int, optional
        The maximum depth of parent or child processes to extract from the data
        (The default is 4)
    processes: pd.DataFrame, optional
        Dataframe of processes to generate tree for

    Returns
    -------
    pd.DataFrame
        The formatted process tree data

    """
    # Superceded by process_tree_utils module
    del branch_depth, processes
    return build_process_tree(audit_data)


def cluster_auditd_processes(audit_data: pd.DataFrame, app: str = None) -> pd.DataFrame:
    """
    Clusters process data into specific processes.

    Parameters
    ----------
    audit_data : pd.DataFrame
        The Audit data containing process creation events
    app: str, optional
        The name of a specific app you wish to cluster

    Returns
    -------
    pd.DataFrame
        Details of the clustered process

    """
    if app is not None:
        processes = audit_data[audit_data["exe"].str.contains(app, na=False)]
    else:
        processes = audit_data
    processes = processes.rename(
        columns={
            "acct": "SubjectUserName",
            "uid": "SubjectUserSid",
            "user": "SubjectUserName",
            "ses": "SubjectLogonId",
            "pid": "NewProcessId",
            "exe": "NewProcessName",
            "ppid": "ProcessId",
            "cmdline": "CommandLine",
        }
    )
    req_cols = [
        "cwd",
        "SubjectUserName",
        "SubjectUserSid",
        "SubjectUserName",
        "SubjectLogonId",
        "NewProcessId",
        "NewProcessName",
        "ProcessId",
        "CommandLine",
    ]
    for col in req_cols:
        if col not in processes:
            processes[col] = ""

    feature_procs_h1 = add_process_features(input_frame=processes)

    clus_events, _, _ = dbcluster_events(
        data=feature_procs_h1,
        cluster_columns=["pathScore", "SubjectUserSid"],
        time_column="TimeGenerated",
        max_cluster_distance=0.0001,
    )
    (
        clus_events.sort_values("TimeGenerated")[
            [
                "TimeGenerated",
                "LastEventTime",
                "NewProcessName",
                "CommandLine",
                "SubjectLogonId",
                "SubjectUserSid",
                "pathScore",
                "isSystemSession",
                "ProcessId",
                "ClusterSize",
            ]
        ].sort_values("ClusterSize", ascending=True)
    )

    procs = clus_events[
        [
            "TimeGenerated",
            "NewProcessName",
            "CommandLine",
            "NewProcessId",
            "SubjectUserSid",
            "cwd",
            "ClusterSize",
            "ProcessId",
        ]
    ]
    procs = procs.rename(columns={"NewProcessId": "pid", "ProcessId": "ppid"})

    return procs



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for Model class for modelling sessions data."""

from collections import defaultdict
from typing import List, Union, Dict

from .utils.data_structures import Cmd
from .utils import cmds_only, cmds_params_only, cmds_params_values, probabilities
from ...common.exceptions import MsticpyException


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
class Model:
    """Class for modelling sessions data."""

    def __init__(
        self, sessions: List[List[Union[str, Cmd]]], modellable_params: set = None
    ):
        """
        Instantiate the Model class.

        This Model class can be used to model sessions, where each
        session is a sequence of commands. We use a sliding window
        approach to calculate the rarest part of each session. We
        can view the sessions in ascending order of this metric to
        see if the top sessions are anomalous/malicious.

        Parameters
        ----------
        sessions: List[List[Union[str, Cmd]]]
            list of sessions, where each session is a list of either
            strings or a list of the Cmd datatype.

            The Cmd datatype should have "name" and "params" as attributes
            where "name" is the name of the command (string) and "params"
            is either a set of accompanying params or a dict of
            accompanying params and values.

            examples formats of a session:
            1) ['Set-User', 'Set-Mailbox']
            2) [Cmd(name='Set-User', params={'Identity', 'Force'}),
            Cmd(name='Set-Mailbox', params={'Identity', 'AuditEnabled'})]
            3) [Cmd(name='Set-User',
            params={'Identity': 'blahblah', 'Force': 'true'}),
            Cmd(name='Set-Mailbox',
            params={'Identity': 'blahblah', 'AuditEnabled': 'false'})]

        modellable_params: set, optional
            set of params which you deem to have categorical values which are suitable
            for modelling.
            Note this argument will only have an effect if your sessions include commands,
            params and values. If your sessions include commands, params and values and
            this argument is not set, then some rough heuristics will be used to determine
            which params have values which are suitable for modelling.

        """
        if not isinstance(sessions, list):
            raise MsticpyException("`sessions` should be a list")
        if len(sessions) == 0:
            raise MsticpyException("`sessions` should not be an empty list")
        for i, ses in enumerate(sessions):
            if not isinstance(ses, list):
                raise MsticpyException("each session in `sessions` should be a list")
            if len(ses) == 0:
                raise MsticpyException(
                    "session at index {} of `sessions` is empty. Each session "
                    "should contain at least one command".format(i)
                )

        self.start_token = "##START##"
        self.end_token = "##END##"
        self.unk_token = "##UNK##"

        self.sessions = sessions
        self.session_type = None
        self._asses_input()

        # non laplace smoothed counts
        self._seq1_counts = None
        self._seq2_counts = None
        self._param_counts = None
        self._cmd_param_counts = None
        self._value_counts = None
        self._param_value_counts = None

        # laplace smoothed counts
        self.seq1_counts = None
        self.seq2_counts = None
        self.param_counts = None
        self.cmd_param_counts = None
        self.value_counts = None
        self.param_value_counts = None

        self.modellable_params = modellable_params

        self.prior_probs = None
        self.trans_probs = None
        self.param_probs = None
        self.param_cond_cmd_probs = None
        self.value_probs = None
        self.value_cond_param_probs = None

        self.set_params_cond_cmd_probs = dict()  # type: Dict[str, Dict[str, float]]

        self.session_likelihoods = None
        self.session_geomean_likelihoods = None

        self.rare_windows = dict()  # type: Dict[int, list]
        self.rare_window_likelihoods = dict()  # type: Dict[int, list]

        self.rare_windows_geo = dict()  # type: Dict[int, list]
        self.rare_window_likelihoods_geo = dict()  # type: Dict[int, list]

    def train(self):
        """
        Train the model by computing counts and probabilities.

        In particular, computes the counts and probabilities of the commands
        (and possibly the params if provided, and possibly the values if provided)

        """
        self._compute_counts()
        self._laplace_smooth_counts()
        self._compute_probs()

    def compute_scores(self, use_start_end_tokens: bool):
        """
        Compute some likelihood based scores/metrics for each of the sessions.

        In particular, computes the likelihoods and geometric mean of
        the likelihoods for each of the sessions. Also, uses the sliding
        window approach to compute the rarest window likelihoods for each
        of the sessions. It does this for windows of length 2 and 3.

        Note that if we have a session of length k, and we use a sliding
        window of length k+1, then we will end up with np.nan for the
        rarest window likelihood metric for that session.
        However, if `use_start_end_tokens` is set to True, then
        because we will be appending self.end_token to the session,
        the session will be treated as a session of length k+1,
        therefore, we will end up with a non np.nan value for that session.

        Parameters
        ----------
        use_start_end_tokens: bool
            if True, then self.start_token and self.end_token will be
            prepended and appended to each
            of the sessions respectively before the calculations are done.

        """
        if self.prior_probs is None:
            raise MsticpyException(
                "please train the model first before using this method"
            )
        self.compute_likelihoods_of_sessions(use_start_end_tokens=use_start_end_tokens)
        self.compute_geomean_lik_of_sessions()
        self.compute_rarest_windows(
            window_len=2, use_geo_mean=False, use_start_end_tokens=use_start_end_tokens
        )
        self.compute_rarest_windows(
            window_len=3, use_geo_mean=False, use_start_end_tokens=use_start_end_tokens
        )

    def _compute_counts(self):
        """
        Compute all the counts for the model.

        The items we will count depend on the the `session_type` attribute.
        We will compute the individual command and transition command counts.

        If params are provided with the commands, then, in addition,
        we will compute the individual param counts and param conditional
        on the command counts.

        If values are provided with the params, then in addition, we
        will compute the individual value counts and value conditional
        on the param counts. Also, we will use rough heuristics
        to determine which params take categorical values, and hence
        have modellable values.

        """
        if self.session_type is None:
            raise MsticpyException("session_type attribute should not be None")

        if self.session_type == SessionType.cmds_only:
            seq1_counts, seq2_counts = cmds_only.compute_counts(
                sessions=self.sessions,
                start_token=self.start_token,
                end_token=self.end_token,
                unk_token=self.unk_token,
            )
            self._seq1_counts = seq1_counts
            self._seq2_counts = seq2_counts

        elif self.session_type == SessionType.cmds_params_only:
            (
                seq1_counts,
                seq2_counts,
                param_counts,
                cmd_param_counts,
            ) = cmds_params_only.compute_counts(
                sessions=self.sessions,
                start_token=self.start_token,
                end_token=self.end_token,
            )

            self._seq1_counts = seq1_counts
            self._seq2_counts = seq2_counts
            self._param_counts = param_counts
            self._cmd_param_counts = cmd_param_counts

        elif self.session_type == SessionType.cmds_params_values:
            (
                seq1_counts,
                seq2_counts,
                param_counts,
                cmd_param_counts,
                value_counts,
                param_value_counts,
            ) = cmds_params_values.compute_counts(
                sessions=self.sessions,
                start_token=self.start_token,
                end_token=self.end_token,
            )

            if self.modellable_params is None:
                modellable_params = cmds_params_values.get_params_to_model_values(
                    param_counts=param_counts, param_value_counts=param_value_counts
                )
                self.modellable_params = modellable_params

            self._seq1_counts = seq1_counts
            self._seq2_counts = seq2_counts
            self._param_counts = param_counts
            self._cmd_param_counts = cmd_param_counts
            self._value_counts = value_counts
            self._param_value_counts = param_value_counts

    def _laplace_smooth_counts(self):
        """
        Laplace smooth all the counts for the model.

        We do this by adding 1 to all the counts. This is so we shift
        some of the probability mass from the very probable
        commands/params/values to the unseen and very unlikely
        commands/params/values. The `unk_token` means we can handle
        unseen commands, params, values, sequences of commands.

        """
        if self._seq1_counts is None:
            raise MsticpyException("Please run the _compute_counts method first.")

        if self.session_type == SessionType.cmds_only:
            seq1_counts_ls, seq2_counts_ls = cmds_only.laplace_smooth_counts(
                seq1_counts=self._seq1_counts,
                seq2_counts=self._seq2_counts,
                start_token=self.start_token,
                end_token=self.end_token,
                unk_token=self.unk_token,
            )
            self.seq1_counts = seq1_counts_ls
            self.seq2_counts = seq2_counts_ls

        elif self.session_type == SessionType.cmds_params_only:
            (
                seq1_counts_ls,
                seq2_counts_ls,
                param_counts_ls,
                cmd_param_counts_ls,
            ) = cmds_params_only.laplace_smooth_counts(
                seq1_counts=self._seq1_counts,
                seq2_counts=self._seq2_counts,
                param_counts=self._param_counts,
                cmd_param_counts=self._cmd_param_counts,
                start_token=self.start_token,
                end_token=self.end_token,
                unk_token=self.unk_token,
            )

            self.seq1_counts = seq1_counts_ls
            self.seq2_counts = seq2_counts_ls
            self.param_counts = param_counts_ls
            self.cmd_param_counts = cmd_param_counts_ls

        elif self.session_type == SessionType.cmds_params_values:
            (
                seq1_counts_ls,
                seq2_counts_ls,
                param_counts_ls,
                cmd_param_counts_ls,
                value_counts_ls,
                param_value_counts_ls,
            ) = cmds_params_values.laplace_smooth_counts(
                seq1_counts=self._seq1_counts,
                seq2_counts=self._seq2_counts,
                param_counts=self._param_counts,
                cmd_param_counts=self._cmd_param_counts,
                value_counts=self._value_counts,
                param_value_counts=self._param_value_counts,
                start_token=self.start_token,
                end_token=self.end_token,
                unk_token=self.unk_token,
            )
            self.seq1_counts = seq1_counts_ls
            self.seq2_counts = seq2_counts_ls
            self.param_counts = param_counts_ls
            self.cmd_param_counts = cmd_param_counts_ls
            self.value_counts = value_counts_ls
            self.param_value_counts = param_value_counts_ls

    def _compute_probs(self):
        """
        Compute all the probabilities for the model.

        The probabilities we compute depends on the `session_type` attribute.
        We will compute the individual command and transition
        command probabilities.

        If params are provided with the commands, then, in addition,
        we will compute the individual param probabilities and param
        conditional on the command probabilities.

        If values are provided with the params, then in addition,
        we will compute the individual value probabilities and
        value conditional on the param probabilities.

        """
        self._compute_probs_cmds()
        if self.session_type in [
            SessionType.cmds_params_only,
            SessionType.cmds_params_values,
        ]:
            self._compute_probs_params()
        if self.session_type == SessionType.cmds_params_values:
            self._compute_probs_values()

    def compute_setof_params_cond_cmd(self, use_geo_mean: bool):  # noqa: MC0001
        """
        Compute likelihood of combinations of params conditional on the cmd.

        In particular, go through each command from each session and
        compute the probability of that set of params (and values if provided)
        appearing conditional on the command.

        This can help us to identify unlikely combinations of params
        (and values if provided) for each distinct command.

        Note, this method is only available if each session is a list
        of the Cmd datatype. It will result in an Exception if you
        try and use it when each session is a list of strings.

        Parameters
        ----------
        use_geo_mean: bool
            if True, then the probabilities will be raised to
            the power of (1/K)

            case1: we have only params:
                Then K is the number of distinct params which appeared
                for the given cmd across all the sessions.
            case2: we have params and values:
                Then K is the number of distinct params which appeared
                for the given cmd across all the sessions + the number
                of values which we included in the modelling for this cmd.

        """
        if self.param_probs is None:
            raise MsticpyException(
                "please train the model first before using this method"
            )

        if self.session_type is None:
            raise MsticpyException("session_type attribute should not be None")

        if self.session_type == SessionType.cmds_only:
            raise MsticpyException(
                'this method is not available for your type of input data "sessions"'
            )
        if self.session_type == SessionType.cmds_params_only:
            result = defaultdict(lambda: defaultdict(lambda: 0))
            for ses in self.sessions:
                for cmd in ses:
                    c_name = cmd.name
                    params = cmd.params
                    prob = cmds_params_only.compute_prob_setofparams_given_cmd(
                        cmd=c_name,
                        params=params,
                        param_cond_cmd_probs=self.param_cond_cmd_probs,
                        use_geo_mean=use_geo_mean,
                    )
                    result[c_name][tuple(params)] = prob
            self.set_params_cond_cmd_probs = result
        else:
            result = defaultdict(lambda: defaultdict(lambda: 0))
            for ses in self.sessions:
                for cmd in ses:
                    c_name = cmd.name
                    params = cmd.params
                    pars = set(cmd.params.keys())
                    intersection_pars = pars.intersection(self.modellable_params)
                    key = set()
                    for par in pars:
                        if par in intersection_pars:
                            key.add("{} --- {}".format(par, params[par]))
                        else:
                            key.add(par)
                    prob = cmds_params_values.compute_prob_setofparams_given_cmd(
                        cmd=c_name,
                        params_with_vals=params,
                        param_cond_cmd_probs=self.param_cond_cmd_probs,
                        value_cond_param_probs=self.value_cond_param_probs,
                        modellable_params=self.modellable_params,
                        use_geo_mean=use_geo_mean,
                    )
                    result[c_name][tuple(key)] = prob
            self.set_params_cond_cmd_probs = result

    def compute_likelihoods_of_sessions(self, use_start_end_tokens: bool = True):
        """
        Compute the likelihoods for each of the sessions.

        Note: If the lengths (number of commands) of the sessions vary a lot,
        then you may not be able to fairly compare the likelihoods between a
        long session and a short session. This is because longer sessions
        involve multiplying more numbers together which are between 0 and 1.
        Therefore the length of the session will be negatively correlated with
        the likelihoods. If you take the geometric mean of the likelihood, then
        you can compare the likelihoods more fairly across different session
        lengths

        Parameters
        ----------
        use_start_end_tokens: bool
            if True, then `start_token` and `end_token` will be prepended
            and appended to the session respectively before the calculations
            are done

        """
        if self.prior_probs is None:
            raise MsticpyException(
                "please train the model first before using this method"
            )

        result = []

        for sess in self.sessions:
            if self.session_type == SessionType.cmds_only:
                tmp = cmds_only.compute_likelihood_window(
                    window=sess,
                    prior_probs=self.prior_probs,
                    trans_probs=self.trans_probs,
                    use_start_token=use_start_end_tokens,
                    use_end_token=use_start_end_tokens,
                    start_token=self.start_token,
                    end_token=self.end_token,
                )
            elif self.session_type == SessionType.cmds_params_only:
                tmp = cmds_params_only.compute_likelihood_window(
                    window=sess,
                    prior_probs=self.prior_probs,
                    trans_probs=self.trans_probs,
                    param_cond_cmd_probs=self.param_cond_cmd_probs,
                    use_start_token=use_start_end_tokens,
                    use_end_token=use_start_end_tokens,
                    start_token=self.start_token,
                    end_token=self.end_token,
                )
            else:
                tmp = cmds_params_values.compute_likelihood_window(
                    window=sess,
                    prior_probs=self.prior_probs,
                    trans_probs=self.trans_probs,
                    param_cond_cmd_probs=self.param_cond_cmd_probs,
                    value_cond_param_probs=self.value_cond_param_probs,
                    modellable_params=self.modellable_params,
                    use_start_token=use_start_end_tokens,
                    use_end_token=use_start_end_tokens,
                    start_token=self.start_token,
                    end_token=self.end_token,
                )

            result.append(tmp)

        self.session_likelihoods = result

    def compute_geomean_lik_of_sessions(self):
        """
        Compute the geometric mean of the likelihood for each of the sessions.

        This is done by raising the likelihood of the session to the power of
        (1 / k) where k is the length of the session.

        Note: If the lengths (number of commands) of the sessions vary a lot,
        then you may not be able to fairly compare the likelihoods between a
        long session and a short session. This is because longer sessions
        involve multiplying more numbers together which are between 0 and 1.
        Therefore the length of the session will be negatively correlated with
        the likelihoods. If you take the geometric mean of the likelihood, then
        you can compare the likelihoods more fairly across different session
        lengths.

        """
        if self.session_likelihoods is None:
            self.compute_likelihoods_of_sessions()
        result = []
        for i in range(len(self.sessions)):
            lik = self.session_likelihoods[i]
            k = len(self.sessions[i])
            result.append(lik ** (1 / k))

        self.session_geomean_likelihoods = result

    def compute_rarest_windows(
        self,
        window_len: int,
        use_start_end_tokens: bool = True,
        use_geo_mean: bool = False,
    ):
        """
        Find the rarest window and corresponding likelihood for each session.

        In particular, uses a sliding window approach to find the rarest window
        and corresponding likelihood for that window for each session.

        If we have a long session filled with benign activity except for a small
        window of suspicious behaviour, then this approach should be able to
        identity the session as anomalous. This approach should be more
        effective than simply taking the geometric mean of the full session
        likelihood. This is because the small window of suspicious behaviour
        might get averaged out by the majority benign behaviour in the session
        when using the geometric mean approach.

        Note that if we have a session of length k, and we use a sliding window
        of length k+1, then we will end up with np.nan for the rarest window
        likelihood metric for that session. However, if `use_start_end_tokens`
        is set to True, then because we will be appending self.end_token to the
        session, the session will be treated as a session of length k+1,
        therefore, we will end up with a non np.nan value.

        Parameters
        ----------
        window_len: int
            length of sliding window for likelihood calculations
        use_start_end_tokens: bool
            if True, then `start_token` and `end_token` will be prepended
            and appended to each
            session respectively before the calculations are done
        use_geo_mean: bool
            if True, then each of the likelihoods of the sliding windows
            will be raised to the power
            of (1/`window_len`)

        """
        if self.prior_probs is None:
            raise MsticpyException(
                "please train the model first before using this method"
            )

        if self.session_type == SessionType.cmds_only:
            rare_tuples = [
                cmds_only.rarest_window_session(
                    session=ses,
                    prior_probs=self.prior_probs,
                    trans_probs=self.trans_probs,
                    window_len=window_len,
                    use_start_end_tokens=use_start_end_tokens,
                    start_token=self.start_token,
                    end_token=self.end_token,
                    use_geo_mean=use_geo_mean,
                )
                for ses in self.sessions
            ]
        elif self.session_type == SessionType.cmds_params_only:
            rare_tuples = [
                cmds_params_only.rarest_window_session(
                    session=ses,
                    prior_probs=self.prior_probs,
                    trans_probs=self.trans_probs,
                    param_cond_cmd_probs=self.param_cond_cmd_probs,
                    window_len=window_len,
                    use_start_end_tokens=use_start_end_tokens,
                    start_token=self.start_token,
                    end_token=self.end_token,
                    use_geo_mean=use_geo_mean,
                )
                for ses in self.sessions
            ]
        else:
            rare_tuples = [
                cmds_params_values.rarest_window_session(
                    session=ses,
                    prior_probs=self.prior_probs,
                    trans_probs=self.trans_probs,
                    param_cond_cmd_probs=self.param_cond_cmd_probs,
                    value_cond_param_probs=self.value_cond_param_probs,
                    modellable_params=self.modellable_params,
                    window_len=window_len,
                    use_start_end_tokens=use_start_end_tokens,
                    start_token=self.start_token,
                    end_token=self.end_token,
                    use_geo_mean=use_geo_mean,
                )
                for ses in self.sessions
            ]

        if use_geo_mean:
            self.rare_windows_geo[window_len] = [rare[0] for rare in rare_tuples]
            self.rare_window_likelihoods_geo[window_len] = [
                rare[1] for rare in rare_tuples
            ]
        else:
            self.rare_windows[window_len] = [rare[0] for rare in rare_tuples]
            self.rare_window_likelihoods[window_len] = [rare[1] for rare in rare_tuples]

    def _compute_probs_cmds(self):
        """Compute the individual and transition command probabilties."""
        if self.seq1_counts is None:
            raise MsticpyException("seq1_counts attribute should not be None")
        if self.seq2_counts is None:
            raise MsticpyException("seq2_counts attribute should not be None")

        prior_probs, trans_probs = probabilities.compute_cmds_probs(
            seq1_counts=self.seq1_counts,
            seq2_counts=self.seq2_counts,
            unk_token=self.unk_token,
        )

        self.prior_probs = prior_probs
        self.trans_probs = trans_probs

    def _compute_probs_params(self):
        """Compute the individual param probs and param conditional on command probs."""
        if self.param_counts is None:
            raise MsticpyException("param_counts attribute should not be None")
        if self.cmd_param_counts is None:
            raise MsticpyException("cmd_param_counts attribute should not be None")

        param_probs, param_cond_cmd_probs = probabilities.compute_params_probs(
            param_counts=self.param_counts,
            cmd_param_counts=self.cmd_param_counts,
            seq1_counts=self.seq1_counts,
            unk_token=self.unk_token,
        )

        self.param_probs = param_probs
        self.param_cond_cmd_probs = param_cond_cmd_probs

    def _compute_probs_values(self):
        """Compute the individual value probs and value conditional on param probs."""
        if self.value_counts is None:
            raise MsticpyException("value_counts attribute should not be None")
        if self.param_value_counts is None:
            raise MsticpyException("param_value_counts attribute should not be None")

        value_probs, value_cond_param_probs = probabilities.compute_values_probs(
            value_counts=self.value_counts,
            param_value_counts=self.param_value_counts,
            unk_token=self.unk_token,
        )

        self.value_probs = value_probs
        self.value_cond_param_probs = value_cond_param_probs

    def _asses_input(self):
        """
        Determine what type of sessions we have.

        In particular, assess the input `self.sessions` to see whether each
        session is a list of strings, or list of the Cmd datatype. And if each
        session is a list of the Cmd datatype, it will assess whether the params
        attribute of the Cmd datatype is a set or a dict.

        """
        session = self.sessions[0]
        cmd = session[0]
        if isinstance(cmd, str):
            self.session_type = SessionType.cmds_only
        elif self._check_cmd_type():
            if isinstance(cmd.params, set):
                self.session_type = SessionType.cmds_params_only
            elif isinstance(cmd.params, dict):
                self.session_type = SessionType.cmds_params_values
            else:
                raise MsticpyException(
                    "Params attribute of Cmd data structure should "
                    + "be either a set or a dict"
                )
        else:
            raise MsticpyException(
                "Each element of 'sessions' should be a list of either "
                + "strings, or Cmd data types"
            )

    def _check_cmd_type(self):
        """Check whether the Cmd datatype has the expected attributes."""
        session = self.sessions[0]
        cmd = session[0]
        if "name" in dir(cmd) and "params" in dir(cmd):
            return True
        return False


class SessionType:
    """Class for storing the types of accepted sessions."""

    cmds_only = "cmds_only"
    cmds_params_only = "cmds_params_only"
    cmds_params_values = "cmds_params_values"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for SecurityEvent class."""
from typing import List, Dict, Any

import pandas as pd

from .entityschema import Entity, Host, Process, Account, IpAddress, HostLogonSession
from .security_base import SecurityBase
from ..common.utility import export
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class SecurityEvent(SecurityBase):
    """SecurityEvent class."""

    def __init__(self, src_row: pd.Series = None):
        """
        Instantiate new instance of SecurityEvent.

            :param src_row: Pandas series containing single security event
        """
        self._source_data = src_row

        super().__init__(src_row=src_row)

        self._extract_entities(src_row)
        self._find_os_family()

    # Properties
    @property
    def entities(self) -> List[Entity]:
        """
        Return the list of entities extracted from the event.

        Returns
        -------
        List[Entity]
            The list of entities extracted from the event.

        """
        return list(self._entities)

    @property
    def query_params(self) -> Dict[str, Any]:
        """
        Query parameters derived from alert.

        Returns
        -------
            Dict[str, Any]
                Dictionary of parameter names

        """
        return super().query_params

    # index operator override
    def __getattr__(self, name):
        """Return the value of the named property 'name'."""
        if name is not None and name in self._source_data:
            return self._source_data[name]
        return None

    def _extract_entities(self, src_row):
        if "EventID" in src_row:
            host = Host(src_event=src_row)
            self._entities.append(host)
            event_id = str(src_row["EventID"])
            if event_id == "4688":
                event_proc = Process(src_event=src_row, role="new")
                self._entities.append(event_proc)
                event_proc["Host"] = host
                if "ParentProcess" in event_proc:
                    self._entities.append(event_proc.ParentProcess)
                    if "ImageFile" in event_proc.ParentProcess:
                        self._entities.append(event_proc.ParentProcess.ImageFile)
                logon_session = HostLogonSession(src_event=src_row)
                logon_session.Host = host
                if "Account" in event_proc:
                    logon_session.Account = event_proc.Account
                    event_proc.Account.Host = host
                    self._entities.append(event_proc.Account)
                self._entities.append(logon_session)
                if "ImageFile" in event_proc:
                    self._entities.append(event_proc.ImageFile)

            if event_id in ("4624", "4625"):
                subj_account = Account(src_event=src_row, role="subject")
                subj_account.Host = host
                self._entities.append(subj_account)
                tgt_account = Account(src_event=src_row, role="target")
                tgt_account.Host = host
                self._entities.append(tgt_account)
                self._entities.append(IpAddress(src_event=src_row))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Settings provider for secrets."""
import re
from functools import partial
from typing import Any, Callable, Dict, Optional, Set, Tuple

import keyring
from keyring.errors import KeyringError, KeyringLocked

from .._version import VERSION
from . import pkg_config as config
from .exceptions import MsticpyKeyVaultConfigError
from .keyvault_client import BHKeyVaultClient
from .keyvault_settings import KeyVaultSettings
from .utility import export

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class KeyringClient:
    """Keyring client wrapper."""

    def __init__(self, name: str = "key-cache", debug: bool = False):
        """
        Initialize the keyring client.

        Parameters
        ----------
        name : str, optional
            Name of the credential group, by default "system"
        debug : bool, optional
            Output debug info, by default False

        """
        self.debug = debug
        self.keyring = name
        self._secret_names: Set[str] = set()

    def __getitem__(self, key: str):
        """Get key name."""
        cred = self.get_secret(key)
        if cred:
            return cred
        raise KeyError

    def get_secret(self, secret_name: str) -> Any:
        """
        Retrieve a secret from the keyring.

        Parameters
        ----------
        secret_name : str
            Secret name.

        Returns
        -------
        Any
            Secret value.

        """
        if self.debug:
            print(f"Fetching {secret_name} from keyring")
        try:
            secret = keyring.get_password(self.keyring, secret_name)
        except (KeyringError, KeyringLocked):
            if self.debug:
                print(
                    "Keyring error retrieving credentials",
                    f"for {secret_name} from keyring {self.keyring}",
                )
        if not secret:
            if self.debug:
                print(
                    "No credentials", f"for {secret_name} from keyring {self.keyring}"
                )
        return secret

    def set_secret(self, secret_name: str, secret_value: Any):
        """
        Set a secret in the keyring group.

        Parameters
        ----------
        secret_name : str
            Name of the secret
        secret_value : Any
            Secret value

        """
        if self.debug:
            print(f"Saving {secret_name} to keyring {self.keyring}")
        self._secret_names.add(secret_name)
        keyring.set_password(self.keyring, secret_name, secret_value)


@export
class SecretsClient:
    """Secrets client - manages keyvault and keyring secrets."""

    def __init__(self, tenant_id: str = None, use_keyring: bool = True):
        """
        Initialize SecretsClient instance.

        Parameters
        ----------
        tenant_id : str, optional
            TenantID, by default None
        use_keyring : bool, optional
            If True use keyring to cache secrets, by default True

        Raises
        ------
        MsticpyKeyVaultConfigError
            Missing or invalid configuration settings.

        Notes
        -----
        Requires KeyVault settings to be defined in msticpyconfig.yaml

        """
        self._kv_settings = KeyVaultSettings()

        self.tenant_id = tenant_id or self._kv_settings.get("tenantid")
        if not self.tenant_id:
            raise MsticpyKeyVaultConfigError(
                "Could not get TenantId from function parameters or configuration.",
                "Please add this to the KeyVault section of msticpyconfig.yaml",
                title="missing tenant ID value.",
            )
        self.kv_secret_vault: Dict[str, str] = {}
        self.kv_vaults: Dict[str, BHKeyVaultClient] = {}
        self._use_keyring = use_keyring or self._kv_settings.get("UseKeyring", False)
        if self._use_keyring:
            self._keyring_client = KeyringClient("Providers")

    def get_secret_accessor(self, setting_path: str) -> Callable[[], Any]:
        """
        Return accessor function for a secret.

        Parameters
        ----------
        setting_path : str
            The msticpy configuration path (dot-separated)

        Returns
        -------
        Callable[[None], Any]
            Accessor function for the secret value.

        """
        vault_name, secret_name = self._get_kv_vault_and_name(setting_path)
        if vault_name is None or secret_name is None:
            return lambda: secret_name if secret_name else ""
        return self._get_secret_func(secret_name, vault_name)

    def _add_key_vault(self, vault_name: str, secret_name: str):
        """Add the KeyVault instance responsible for storing `secret_name`."""
        vault = self.kv_vaults.get(vault_name)
        if not vault:
            vault = BHKeyVaultClient(self.tenant_id, vault_name=vault_name)
            self.kv_vaults[vault_name] = vault
        self.kv_secret_vault[secret_name] = vault_name

    @staticmethod
    def format_kv_name(setting_path):
        """Return normalized name for use as a KeyVault secret name."""
        return re.sub("[^0-9a-zA-Z-]", "-", setting_path)

    def _get_kv_vault_and_name(
        self, setting_path: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return the vault and secret name for a config path."""
        setting_item = config.get_config(setting_path)

        if not isinstance(setting_item, dict):
            return None, str(setting_item)
        if "KeyVault" in setting_item:
            kv_val = setting_item.get("KeyVault")
            def_vault_name = self._kv_settings.get("VaultName")
            if not kv_val or kv_val.casefold() == "default":
                # If no value, get the default VaultName from settings
                # and use the setting path as the secret name
                if not def_vault_name:
                    raise ValueError("No VaultName defined in KeyVault settings.")
                secret_name = self.format_kv_name(setting_path)
                return def_vault_name, secret_name
            if "/" in kv_val:
                # '/' delimited string means VaultName/Secret
                vault_name, secret_name = kv_val.split("/")
                return vault_name, self.format_kv_name(secret_name)
            if not def_vault_name:
                raise MsticpyKeyVaultConfigError(
                    "Check that you have specified the right value for VaultName"
                    + " in your configuration",
                    f"No VaultName defined in KeyVault settings for {setting_path}.",
                    title="Key Vault vault name not found.",
                )
            # If there is a single string - take that as the secret name
            return def_vault_name, self.format_kv_name(kv_val)
        return None, None

    def _get_secret_func(self, secret_name: str, vault_name: str) -> Callable[[], Any]:
        """Return a func to access a secret."""
        if self._use_keyring:
            if self._keyring_client.get_secret(secret_name):
                return self._create_secret_func(self._keyring_client, secret_name)

        # If the secret is not in keyring, get the vault holding this secret
        if not self.kv_secret_vault.get(secret_name):
            self._add_key_vault(secret_name=secret_name, vault_name=vault_name)

        vault = self.kv_vaults[vault_name]
        if self._use_keyring:
            # store the secret in keyring and return an accessor
            # to the keyring value.
            self._keyring_client.set_secret(secret_name, vault.get_secret(secret_name))
            return self._create_secret_func(self._keyring_client, secret_name)
        # if not using Keyring - return a KeyVault accessor
        return self._create_secret_func(vault, secret_name)

    @staticmethod
    def _create_secret_func(secret_store, secret_name):
        return partial(secret_store.get_secret, secret_name=secret_name)

    @staticmethod
    def read_secret(secret_object: Any) -> Any:
        """
        Return the secret value.

        Parameters
        ----------
        secret_object : Any
            If it is a func, call and return the return value
            of that func. Otherwise just return the object.

        Returns
        -------
        Any
            The secret value

        """
        if callable(secret_object):
            return secret_object()
        return secret_object
